/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <getopt.h>

#include "common.h"
#include "options.h"
#include "log.h"
#include "util.h"
#include "core.h"
#include "location.h"
#include "remstore.h"

#define ARG_COUNT 3
static inline void usage( char* prog )
{
    printf( "Usage: %s [args] [options]\n"
            "Arguments:\n"
            "-s statefile\t"   "Read client state from statefile.cst\n"
            "Options:\n"
		    "-l logdir     \t" "Directory where output logs and data are stored\n"
            "-r resourcedir\t" "Directory containing resources (e.g. options.adeona)\n"
            "-o optsfile\t"    "Use options specified in file optsfile\n"
            "\n", prog );
}

static int update_location_cache( state_t* state, update_t* update ) 
{
    if( state == NULL || update == NULL ) {
        error( "invalid input" );
        return FAILURE;
    }

    location_alloc( &update->loc );

    debug( "======== start location finder =========" );
    if( location_get_data( &update->loc ) == FAILURE ) {
        error( "could not get location data" );
        location_free( &update->loc );
        return FAILURE;
    }
    
    log_msg_func( ADEONA_DEBUG, location_printdata_compact, &update->loc );

    debug( "========= end location finder ==========" );

    if( ctxt_cache_update( state->cacheseed, state->cache, update ) == FAILURE ) {
		error( "could not update ctxt cache" );
        location_free( &update->loc );
		return FAILURE;
	}

    location_free( &update->loc );

    return SUCCESS;
}


static void nonpanic_clientloop( char* cstatefilename, state_t* state  )
{
    assert( cstatefilename != NULL && state != NULL );
    
    struct in_addr internet_check;
    update_t update;
    update.skipped = 0;
    while( TRUE ) {
        int catchup_skipped = 0;
        if( sched_catchup( state, &catchup_skipped ) == FAILURE ) {
            error( "could not catchup state" ); //TODO: this is fatal, no?
        }
        update.skipped += catchup_skipped;
        
        while( sched_sleep( state ) == CONTINUE ) {
            // only slept MAX_SLEEPTIME seconds, check location
            // Here we only update if in fact we can unambiguously determine
            // if location changed. This prevents updating every MAX_SLEEPTIME seconds if
            // for some reason we can't determine external or internal ip address 
            // at this location.
            if( location_checkchanged() == SUCCESS ) {
                debug( "location changed, updating cache" );
                update_location_cache( state, &update );
                if( ismsbset( update.skipped ) == FAILURE )
                    update.skipped = flipmsb( update.skipped ); // Recorded this skip, so flip msb to signal that next 
                                                                // cache element will also be from this state
            }
        }
        
        // After waking up, we check for Internet
        // If there's no connection, we skip this update
        if( get_internalip( &internet_check ) == FAILURE ) {
            debug( "========== START SKIPPED UPDATE ==========" );
            log_msg_bin( ADEONA_CRYPTO, state->currentseed, sizeof( prgkey_t ), "state currentseed:" );
            debug( "state updatetime: %s", time2string( state->updatetime, TRUE ) );
            
            if( sched_nextstate( state ) == FAILURE ) {
                error( "could not move to next state" ); //TODO: also fatal?
            }
            update.skipped = 1;
            
            debug( "========== END SKIPPED UPDATE ==========" );
            
            continue;
        }
        
        // Update cache if location change returns FAILURE or SUCCESS 
        // (i.e. only DON'T update if we can decidedly determine that we
        // haven't changed locations.)
        // OR if we have yet to report a location for this update time
        // period (this maintains continuity of the skipped list)
        // 
        if( location_checkchanged() != CONTINUE || ismsbset( update.skipped ) == FAILURE ) {
            update_location_cache( state, &update );
        }
        
        debug( "========== START STATE UPDATE ==========" );
        log_msg_bin( ADEONA_CRYPTO, state->currentseed, sizeof( prgkey_t ), "state currentseed:" );
        debug( "state updatetime: %s", time2string( state->updatetime, TRUE ) );
       
        if( update_encapsulate( state ) != FAILURE 
            && update_interrupt_protection( cstatefilename, state ) != FAILURE ) {

		    if( remstore_connect() == SUCCESS ) {
				int i;
                int goodput = opts.num_dhtkeys_totry;
				for( i=0; i<opts.num_dhtkeys_totry; i++ ) {
					
                    log_msg_bin( ADEONA_DEBUG, state->sd->dhtkeys[i], sizeof( dhtkey_t ), "INSERT to dht key %d", i );
                    log_msg_bin( ADEONA_DEBUG, (uchar*)state->cache, CTXT_QUEUE_LEN, "inserting ctxtcache:" );
                    
                    if( remstore_isconnected() == FALSE ) {
                        debug( "connection lost in previous key insert, reconnecting..." );
                        if( remstore_connect() == FALSE ) {
                            error( "could not connect for current key" );
                            goodput--;
                            continue;
                        }
                    }

					if( remstore_put( state->sd->dhtkeys[i], (uchar*)state->cache, CTXT_QUEUE_LEN ) == FAILURE ) {
						error( "could not insert to dht" );
                        goodput--;
					}
				}
                
                debug( "%d successful inserts out of %d", goodput, opts.num_dhtkeys_totry );

				remstore_disconnect();
			}
			else {
                error( "could not connect to remote storage server" );
            }
		}    
		else {
			error( "could not perform update, failed to encapsulate or protect from interrupts" );
		}

        if( sched_nextstate( state ) == FAILURE ) {
            error( "could not move to next state" ); // if we can't advance the state, we'll end up reusing it which not good, fatal?
		}
        update.skipped = 1;
        
        debug( "==========  END STATE UPDATE  ==========" );
    } // end while
}

int main( int argc, char** argv )
{
    if( argc < ARG_COUNT ) {
        usage( argv[0] );
        exit( EXIT_FAILURE );
    }

    int opt;
	char* logdir = NULL;
    char* resourcedir = NULL;
    char* cstatefilename = NULL;
    
    while( ( opt = getopt( argc, argv, "l:r:s:o:h" ) ) != -1 ) {
        switch( opt ) {
            case 's' : cstatefilename = optarg; break;
            case 'l' : logdir = optarg; break;
            case 'r' : resourcedir = optarg; break;
            case 'h' : /* fall through */
            default :
                usage( argv[0] );
                exit( EXIT_SUCCESS );
            break;
        }
    }
    
    if( cstatefilename == NULL  ) {
        usage( argv[0] );
        exit( EXIT_FAILURE );
    }
    
	setup_logdir( logdir, "adeona_client_log.txt" );
	setup_resources( resourcedir, RES_OPTIONS | RES_GATEWAYS | RES_LANDMARKS | RES_WEBCAM );

    state_t* state = state_alloc();
    assert( state != NULL );
        
    if( state_load_filename( cstatefilename, state, NULL, 0, SC_ALL ) == FAILURE ) {
		state_free( state );
		fatal( "could not load state from %s", cstatefilename );
	}
	
#if defined( _ADEONA_BUILD_RELEASE_ ) 
    info( "***** This log file is intentionally sparse to preserve the privacy of past locations from a thief. *****" );
#endif
    debug( "================ ADEONA CLIENT STARTING UP ================" );
    nonpanic_clientloop( cstatefilename, state );
    
	state_free( state );
    
    exit( EXIT_SUCCESS );
}
