/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include "common.h"
#include "log.h"
#include "util.h"

static void test_hexconversion( void )
{
    int i;
    int len = 16;
    uchar allones[len];
    uchar hexbuff[len*2];
    
    memset( allones, 0xFF, sizeof( allones ) );
    memset( hexbuff, 0x00, sizeof( hexbuff ) );
    
    bin2hex( hexbuff, sizeof( hexbuff ), allones, sizeof( allones ) );
    memset( allones, 0x00, sizeof( allones ) );
    for( i=0; i<sizeof( allones ); i++ ) {
        assert( allones[i] == 0 );
    }

    hex2bin( hexbuff, sizeof( hexbuff ), allones, sizeof( allones ) );
    for( i=0; i<sizeof( allones ); i++ ) {
        assert( allones[i] == 0xFF );
    }
    printout( "%s tests passed\n", __FUNCTION__ );
}

static void test_filerw_chars( void )
{
    char* filename = "savefile.txt";
    FILE* file = fopen( filename, "w+" );
    assert( file != NULL );
    
    int len = 13;
    int half = len/2;
    int otherhalf = ( len%2 == 0 ) ? half:( half+1 );
    uchar savethis[len];
    memset( savethis, 0xAA, half );
    memset( &savethis[half], 0xBB, otherhalf );

    fprintln_bin2hex( file, savethis, sizeof( savethis ) );
    rewind( file );

    uchar loadthis[len];
    memset( loadthis, 0x00, sizeof( loadthis ) );
    
    freadln_hex2bin( file, loadthis, sizeof( loadthis ) );

    int i;
    for( i=0; i<len; i++ ) {
        assert( savethis[i] == loadthis[i] );
    }
    
    printout( "%s test passed\n", __FUNCTION__ );
    system( "rm savefile.txt" );
}

static void test_dataserialize( void )
{
    char* filename = "dataserialize.txt";
    FILE* file = fopen( filename, "w+" );
    assert( file != NULL );
    
    int theint = -1430532899;
    int loadint;
    short theshort = -21829;
    short loadshort;
    long thelong = -1430532899;
    long loadlong;
    double thedouble = -12.051984;
    double loaddouble;
    time_t thetime = -1430532899;
    time_t loadtime;
    
    int errorint = 1;

    fprintln_data2hex( file, &theint, sizeof( int ) );
    fprintln_data2hex( file, &theshort, sizeof( short ) );
    fprintln_data2hex( file, &thelong, sizeof( long ) );
    fprintln_data2hex( file, &thedouble, sizeof( double ) );
    fprintln_data2hex( file, &thetime, sizeof( time_t ) );
    fprintf( file, "AABBCCGG\n" );
    fprintf( file, "AABB\n\nCCGG\n" ); //under
    fprintf( file, "AABBCCDDEEFFGG\n" ); //over

    rewind( file );

    freadln_hex2data( file, &loadint, sizeof( int ) );
    assert( theint == loadint );
    
    freadln_hex2data( file, &loadshort, sizeof( short ) );
    assert( theshort == loadshort );
    
    freadln_hex2data( file, &loadlong, sizeof( long ) );
    assert( thelong == loadlong );
    
    freadln_hex2data( file, &loaddouble, sizeof( double ) );
    assert( thedouble == loaddouble );
    
    freadln_hex2data( file, &loadtime, sizeof( time_t ) );
    assert( thetime == loadtime );
    
    freadln_hex2data( file, &errorint, sizeof( int ) );
    assert( errorint == 1 ); 
    
    freadln_hex2data( file, &errorint, sizeof( int ) );
    assert( errorint == 1 );
    
    freadln_hex2data( file, &errorint, sizeof( int ) );
    assert( errorint == 1 );

    printout( "%s tests passed\n", __FUNCTION__ );
    system( "rm dataserialize.txt" );
}

//functional test for endianness stuff
static void load_serialized( char* filename )
{
    FILE* file = fopen( filename, "r" );
    assert( file != NULL );
    
    int theint = -1430532899;
    int loadint;
    short theshort = -21829;
    short loadshort;
    long thelong = -1430532899;
    long loadlong;
    double thedouble = -12.051984;
    double loaddouble;
    time_t thetime = -1430532899;
    time_t loadtime;
    
    freadln_hex2data( file, &loadint, sizeof( int ) );
    freadln_hex2data( file, &loadshort, sizeof( short ) );
    freadln_hex2data( file, &loadlong, sizeof( long ) );
    freadln_hex2data( file, &loaddouble, sizeof( double ) );
    freadln_hex2data( file, &loadtime, sizeof( time_t ) );
    
    assert( theint == loadint );
    assert( theshort == loadshort );
    assert( thelong == loadlong );
    assert( thedouble == loaddouble );
    assert( thetime == loadtime );
    
    printout( "%s tests passed\n", __FUNCTION__ );
}

static void test_genrandom( void )
{
    uint bsize = 1024*1024;
    uchar* rbuff = (uchar*)malloc( bsize );
    
    assert( get_random_bytes( rbuff, bsize ) == SUCCESS );
    
//    int i;
//    for( i=0; i<1000; i++ ) {
//        assert( get_random_bytes( rbuff, bsize ) == SUCCESS );
//    }

    free( rbuff );
    
    printout( "%s tests passed\n", __FUNCTION__ );
}

static void test_timeconversion( void )
{
    time_t t = time( NULL );

    char* tstr1 = time2string( t, TRUE );
    time_t t1 = string2time( tstr1 ); //tstr1 is ptr to static buffer so should use value right away
    
    char* tstr2 = time2string( t, TRUE );
    time_t t2 = string2time( tstr2 );
    
    char* tstr3 = time2string( t, FALSE );
    time_t t3 = string2time( tstr3 );
    
    double diff1 = difftime( t1, t2 );
    double diff2 = difftime( t1, t3 );
    double diff3 = difftime( t2, t3 );
    
    assert( diff1 == 0 );
    assert( diff2 != 0 && diff3 != 0 && diff2 == diff3 );
    
    printout( "%s test passed\n", __FUNCTION__ );

}

static void test_exec_command( void )
{

    char buffer[4096];
    memset( buffer, 0x00, sizeof( buffer ) );
//    char* cmd = "ping -c 10 www.google.com";
    char* cmd = "traceroute -q 1 -m 20 -n www.google.com";
    
    exec_command( cmd, buffer, sizeof( buffer ) );
    printout( "%s\n", buffer );

//    int i;
//    for( i=0; i<1000; i++ ) {
//        exec_command( cmd, buffer, sizeof( buffer ) );
//    }
    
}

static void test_allowed_chars_only( void )
{
    char* allowed = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ *-";
    char* hasbad  = "The quick brown fox jumps + over the lazy dog";
    char* hasbad2  = "The quick brown fox jumps over the lazy dog.";
    char* nobad   = "The quick brown fox jumps over the lazy dog THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG * -- * --";
    char* overflow = (char*)malloc( strlen( nobad ) + 2 ); //what if overflow
    memcpy( overflow, nobad, strlen( nobad ) );
    overflow[strlen( nobad )] = 'a'; //is the allowed char but overflow
    overflow[strlen( nobad ) + 1] = '\0';

    assert( allowed_chars_only( hasbad, strlen( hasbad ), allowed ) == FALSE );
    assert( allowed_chars_only( hasbad2, strlen( hasbad2 ), allowed ) == FALSE );
    assert( allowed_chars_only( nobad, strlen( nobad ), allowed ) == TRUE );
    assert( allowed_chars_only( overflow, strlen( nobad ), allowed ) == FALSE );
    
    free( overflow );

    printout( "%s tests passed\n", __FUNCTION__ );
}

int main( void )
{
    test_hexconversion();        
    test_filerw_chars();
    test_dataserialize();
    test_genrandom();
    test_timeconversion();
    test_allowed_chars_only();
//    load_serialized(); // test loading file saved from different endianness
//    test_exec_command();
    
    exit( EXIT_SUCCESS );
}
