/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <stdarg.h>
#include <unistd.h>

#include "log.h"
#include "util.h"
#include "options.h"

static FILE* _logfile = NULL;
#define set_logfile() { if( _logfile == NULL ) { _logfile = stdout; } }

/* 
 * Adds our disclaimer to all log files
 *
 */ 
static inline void print_disclaimer( void )
{
    fprintf( _logfile,
             "*******************************************************************************" NEWLINE_CHAR 
             "* These results are for informational, research and evaluation purposes only. *" NEWLINE_CHAR
             "* Do not attempt to recover your lost or stolen laptop yourself.              *" NEWLINE_CHAR
             "* If you believe your laptop has been stolen, contact the appropriate         *" NEWLINE_CHAR
             "* law enforcement agency.                                                     *" NEWLINE_CHAR
             "*******************************************************************************" NEWLINE_CHAR NEWLINE_CHAR );
}

static inline void print_timestamp( void )
{
    if( LOG_USETIMESTAMP ) {
        time_t timestamp = time( NULL );
        fprintf( _logfile, "[%s] ", time2string( timestamp, FALSE ) );
//        fprintf( _logfile, "[%d] ", (int)timestamp );
    }
}

int log_open( char* filename )
{
    if( filename != NULL ) {
#if defined( _ADEONA_BUILD_RELEASE_ )
        _logfile = fopen( filename, "w" );
#else
        _logfile = fopen( filename, "a" );
#endif
        if( _logfile == NULL) {
            return FAILURE;
        }
    }
    else {
        _logfile = stdout;
    }

    print_disclaimer();

    return SUCCESS;
}

void log_close( void )
{
    if( _logfile != NULL && _logfile != stdout ) {
        fsync( fileno( _logfile ) );
        fclose( _logfile );
        _logfile = NULL;
    }
    else {
        fprintf( stdout, "warning: cannot close logfile\n" );
    }
}

void log_msg( logdata_t type, char* fmtstr, ... )
{
    set_logfile();
    
    va_list ap;
    
    if( ( type == ADEONA_INFO && LOG_INFO == 1 ) 
        || ( type == ADEONA_DEBUG && LOG_DEBUG == 1 )
        || ( type == ADEONA_CRYPTO && LOG_CRYPTO == 1 ) )
    {

        print_timestamp();
        
        if( type == ADEONA_INFO ) {
            fprintf( _logfile, "info: " );
        }
        else if( type == ADEONA_DEBUG ) {
            fprintf( _logfile, "debug: " );
        }
        else if( type == ADEONA_CRYPTO ) {
            fprintf( _logfile, "crypto: " );
        }
        
        va_start( ap, fmtstr );
	    vfprintf( _logfile, fmtstr, ap );
	    va_end( ap );
	    fprintf( _logfile, NEWLINE_CHAR );
    
        fflush( _logfile );
        fsync( fileno( _logfile ) );
    }
}

void log_msg_bin( logdata_t type, uchar* data, int dlen, char* fmtstr, ... )
{
    set_logfile();
    
    va_list ap;

    if( ( type == ADEONA_INFO && LOG_INFO == 1 ) 
        || ( type == ADEONA_DEBUG && LOG_DEBUG == 1 )
        || ( type == ADEONA_CRYPTO && LOG_CRYPTO == 1 ) )
    {

        print_timestamp();
        
        if( type == ADEONA_INFO ) {
            fprintf( _logfile, "info: " );
        }
        else if( type == ADEONA_DEBUG ) {
            fprintf( _logfile, "debug: " );
        }
        else if( type == ADEONA_CRYPTO ) {
            fprintf( _logfile, "crypto: " );
        }

        va_start( ap, fmtstr );
	    vfprintf( _logfile, fmtstr, ap );
	    va_end( ap );
	    fprintf( _logfile, " " );

        int i;
        for( i=0; i<dlen; i++ ) {
            fprintf( _logfile, "%02X", data[i] );
        }
        fprintf( _logfile, NEWLINE_CHAR );

        fflush( _logfile );
        fsync( fileno( _logfile ) );
    }
}

void log_msg_func( logdata_t type, logfunc_t interpreter, void* data )
{
    set_logfile();

    if( ( type == ADEONA_INFO && LOG_INFO == 1 ) 
        || ( type == ADEONA_DEBUG && LOG_DEBUG == 1 )
        || ( type == ADEONA_CRYPTO && LOG_CRYPTO == 1 ) )
    {
        interpreter( _logfile, data );
        fflush( _logfile );
        fsync( fileno( _logfile ) );
    }
}

void log_err( logdata_t type, const char* file, const char* function, const unsigned long line, char* fmtstr, ... )
{
    set_logfile();
    
    va_list ap;
    
    print_timestamp();

    static const char* _error_message[] = { 
        "warning: ",
        "error: ",
        "fatal error: ",
    };
    
    const char* errmsg = NULL;
    switch( type ) {
        case ADEONA_WARNING :
            errmsg = _error_message[0];
            break;
        case ADEONA_ERROR :
            errmsg = _error_message[1];
            break;
        case ADEONA_FATAL :
            errmsg = _error_message[2];
            break;
        default :
            errmsg = _error_message[0];
    }
    
    if( ( type == ADEONA_WARNING && LOG_WARNING == 1 ) 
        || ( type == ADEONA_ERROR && LOG_ERROR == 1 ) )
    {
        fprintf( _logfile, "%s:%lu: %s", function, line, errmsg );
        
        va_start( ap, fmtstr );
        vfprintf( _logfile, fmtstr, ap );
        va_end( ap );
        fprintf( _logfile, NEWLINE_CHAR );
        
        fflush( _logfile );
        fsync( fileno( _logfile ) );
        return;
    }

    //we want to write fatal messages to stdout all the time
    if( type == ADEONA_FATAL && LOG_FATAL == 1 ) {
        fprintf( _logfile, "%s:%lu: %s", function, line, errmsg );
        if( _logfile != stdout ) {
            fprintf( stdout, "%s:%lu: %s", function, line, errmsg );
        }
        
        va_start( ap, fmtstr );
        vfprintf( _logfile, fmtstr, ap );
        if( _logfile != stdout ) {
            vfprintf( stdout, fmtstr, ap );
        }
        va_end( ap );

        fprintf( _logfile, NEWLINE_CHAR );
        if( _logfile != stdout ) {
            fprintf( stdout, NEWLINE_CHAR );
        }
        
        fflush( _logfile );
        fsync( fileno( _logfile ) );
        
        exit( EXIT_FAILURE );
    }
}

/*
 * printf wrapper.
 *
 */ 
void printout( char* fmtstr, ... )
{
    if( opts.have_console == TRUE ) {
        va_list ap;
        
        va_start( ap, fmtstr );
        vfprintf( stdout, fmtstr, ap );
        va_end( ap );
        
        fflush( stdout );
    }
    return;
}


