/*
 * Copyright (c) 2007-2008 University of Washington
 *
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include "common.h"
#include "log.h"
#include "options.h"
#include "util.h"
#include "landmarks.h"
#include "gateways.h"

#define STR_OPTION_FMT "%[^\r\n]1024s"
#define INT_OPTION_FMT "%d"
#define FLT_OPTION_FMT "%f"

#define	OPTION_LOG_FILENAME       "LOG_FILENAME"
#define OPTION_CTXT_QUEUE_SIZE    "CTXT_QUEUE_SIZE"
#define OPTION_SCHED_LAMBDA       "SCHED_LAMBDA"
#define OPTION_SERVER_PORT        "SERVER_PORT"
#define OPTION_SERVER_NAME        "SERVER_NAME"
#define OPTION_NUM_DHTKEYS_TOTRY  "NUM_DHTKEYS_TOTRY"
#define OPTION_GATEWAYS_FILENAME  "GATEWAYS_FILENAME"
#define OPTION_USE_TRACEROUTES    "USE_TRACEROUTES"
#define OPTION_LANDMARKS_FILENAME "LANDMARKS_FILENAME"
#define OPTION_MAX_SLEEPTIME      "MAX_SLEEPTIME"
#define OPTION_USE_WEBCAMIMG      "USE_WEBCAMIMG"
#define OPTION_WEBCAMPROG_PATH    "WEBCAMPROG_PATH"
#define OPTION_GWUPDATE_URL       "GWUPDATE_URL"
#define OPTION_USE_GWUPDATE       "USE_GWUPDATE"
#define OPTION_EXTIP_URL          "EXTIP_URL"
#define OPTION_END                "END_OPTIONS"

//NOTE: all strings must be less than 1023 bytes.
#define DEFAULT_RESOURCE_PATH      "resources/"
#define	DEFAULT_LOG_FILENAME       "adeona.log" 
#define DEFAULT_SERVER_PORT        5852
#define DEFAULT_CTXT_QUEUE_SIZE    1
#define DEFAULT_SCHED_LAMBDA       0.0006
#define DEFAULT_SERVER_NAME        "opendht.nyuld.net" 
#define DEFAULT_NUM_DHTKEYS_TOTRY  2
#define DEFAULT_GATEWAYS_FILENAME  "resources/gateways.adeona"
#define DEFAULT_USE_TRACEROUTES    1
#define	DEFAULT_LANDMARKS_FILENAME "resources/landmarks.adeona"
#define	DEFAULT_MAX_SLEEPTIME      300 // 5 minutes
#define DEFAULT_USE_WEBCAMIMG      0
#define DEFAULT_WEBCAMPROG_PATH    "resources/isightcapture.exe"
#define DEFAULT_GWUPDATE_URL       "adeona.cs.washington.edu/gateways.adeona"
#define DEFAULT_EXTIP_URL          "adeona.cs.washington.edu/cgi-bin/ip.cgi"
#define DEFAULT_USE_GWUPDATE       1 

/*
 * Globally-accessible options with some sane defaults
 *
 */
options_t opts;

static inline void init_opts( void )
{
    static int fRunAlready = 0;
    if( fRunAlready == 0 ) {
    	memset( &opts, 0x00, sizeof( options_t ) );
	    opts.have_console = FALSE;
        fRunAlready = 1;
    }
}

static inline void load_default_logfile( void )
{
    if( log_open( DEFAULT_LOG_FILENAME ) == FAILURE ) {
	    error( "Could not open default log file" );
    }
    else {
        info( "==========================================" );
        info( "========== LOADING OPTIONS FILE ==========" );
        fflush( NULL );
    }
}

static inline void load_default_lmfile( void )
{
    //info( "Could not load specified landmarks file, using default" );
    if( landmarks_load( DEFAULT_LANDMARKS_FILENAME ) == FAILURE ) {
        error( "Could not open default landmarks file" );
    }
    else {
        info( "LANDMARKS_FILENAME = %s", DEFAULT_LANDMARKS_FILENAME );
    }
}

static inline void load_default_gwfile( void )
{
    //info( "Could not load specified gateways file, using default" );
    if( gateways_load( DEFAULT_GATEWAYS_FILENAME ) == SUCCESS ) {
        info( "GATEWAYS_FILENAME = %s", DEFAULT_GATEWAYS_FILENAME );
        memcpy( opts.gateways_filename, DEFAULT_GATEWAYS_FILENAME, strlen( DEFAULT_GATEWAYS_FILENAME ) );
    }
    else {
        error( "Could not open default gateways file" );
        memset( opts.gateways_filename, 0x00, MAX_STROPTION_LEN );
    }
}

/*
 * So we can load necessary files when the options file is not given 
 */ 
void load_default_options( void )
{
	info( "using default options" );

    init_opts();

    opts.ctxt_queue_size   = DEFAULT_CTXT_QUEUE_SIZE;
    
    opts.sched_lambda      = DEFAULT_SCHED_LAMBDA;
    
    opts.server_port       = DEFAULT_SERVER_PORT;
    
    memcpy( opts.server_name, DEFAULT_SERVER_NAME, strlen( DEFAULT_SERVER_NAME ) );
    //info( "SERVER_NAME = %s", opts.server_name );
    
    opts.num_dhtkeys_totry = DEFAULT_NUM_DHTKEYS_TOTRY;
    
    opts.use_traceroutes   = DEFAULT_USE_TRACEROUTES;
    
    opts.max_sleeptime     = DEFAULT_MAX_SLEEPTIME;

    opts.use_webcamimg     = DEFAULT_USE_WEBCAMIMG;
    
    memcpy( opts.gwupdate_url, DEFAULT_GWUPDATE_URL, strlen( DEFAULT_GWUPDATE_URL ) );
    
    opts.use_gwupdate      =  DEFAULT_USE_GWUPDATE;
    
    memcpy( opts.extip_url, DEFAULT_EXTIP_URL, strlen( DEFAULT_EXTIP_URL ) );
    
}

int write_options_file( FILE* file IN, int fRetrievalOpts IN )
{
    if( file == NULL ) {
        error( "invalid input" );
        return FAILURE;
    }

    if( fRetrievalOpts == 0 ) {
        // Write out all of the options!
        //fprintf( file, "%s = %s\n", OPTION_LOG_FILENAME, opts.log_filenamebase );
        fprintf( file, "%s = %d\n", OPTION_CTXT_QUEUE_SIZE, opts.ctxt_queue_size );
        fprintf( file, "%s = %f\n", OPTION_SCHED_LAMBDA, opts.sched_lambda );
        fprintf( file, "%s = %d\n", OPTION_SERVER_PORT, opts.server_port );
        fprintf( file, "%s = %s\n", OPTION_SERVER_NAME, opts.server_name );
        fprintf( file, "%s = %d\n", OPTION_NUM_DHTKEYS_TOTRY, opts.num_dhtkeys_totry );
        //fprintf( file, "%s = %s\n", OPTION_GATEWAYS_FILENAME, opts.gateways_filename );
        fprintf( file, "%s = %d\n", OPTION_USE_TRACEROUTES, opts.use_traceroutes );
        //fprintf( file, "%s = %s\n", OPTION_LANDMARKS_FILENAME, opts.landmarks_filename );
        //fprintf( file, "%s = %d\n", OPTION_USE_WEBCAMIMG, opts.use_webcamimg );
        //fprintf( file, "%s = %s\n", OPTION_WEBCAMPROG_PATH, opts.webcamprog_path );
        fprintf( file, "%s\n", OPTION_END );
    }
    else {
        // Write out only retrieval-relevant options
        fprintf( file, "%s = %d\n", OPTION_CTXT_QUEUE_SIZE, opts.ctxt_queue_size );
        fprintf( file, "%s = %f\n", OPTION_SCHED_LAMBDA, opts.sched_lambda );
        fprintf( file, "%s = %d\n", OPTION_NUM_DHTKEYS_TOTRY, opts.num_dhtkeys_totry );
        fprintf( file, "%s = %d\n", OPTION_USE_TRACEROUTES, opts.use_traceroutes );
        fprintf( file, "%s = %d\n", OPTION_USE_WEBCAMIMG, opts.use_webcamimg );
        fprintf( file, "%s = %d\n", OPTION_USE_GWUPDATE, opts.use_gwupdate );
        fprintf( file, "%s\n", OPTION_END );
    }
    return SUCCESS;
}

int setup_logdir( char* logdir, char* logname )
{
	if( logname == NULL ) {
		error( "invalid input" );
		return FAILURE;
	}

    init_opts();

	char* filename = NULL;
	if( logdir == NULL ) 
		filename = concatstr( "", logname ); 
    else
		filename = concatstr( logdir, logname );

	if( ( strlen( filename ) < MAX_STROPTION_LEN ) && 
        allowed_chars_only( filename, strlen( filename ), FILENAME_ALLOWED_CHARS ) == TRUE ) {

		if( log_open( filename ) == SUCCESS ) {
            if( logdir != NULL ) {
                memcpy( opts.log_filenamebase, logdir, strlen( logdir ) );
            }

            info( "==========================================" );
			info( "========== LOADING OPTIONS FILE ==========" );
			fflush( NULL );
			free( filename );
			return SUCCESS;
		}
		else {
			info( "could not open log file" );
		}
	}
	free( filename );
    
	info( "failed to load specified log file" ); 
	load_default_logfile();
	
    return FAILURE;
}

int setup_resources( char* resourcedir, int flags )
{
	char path[MAX_STROPTION_LEN];
	char* filename = NULL;

	memset( path, 0x00, MAX_STROPTION_LEN );

	if( resourcedir == NULL ) { 
		memcpy( path, DEFAULT_RESOURCE_PATH, strlen( DEFAULT_RESOURCE_PATH ) );
	}
	else { 
		if( strlen( resourcedir ) > MAX_STROPTION_LEN - strlen( "gateways.adeona" ) ) {
			error( "resourcedir is too long" );
			load_default_options();
			if( flags & RES_GATEWAYS ) load_default_gwfile();
			if( flags & RES_LANDMARKS ) load_default_lmfile();
			return FAILURE;
		}
		memcpy( path, resourcedir, strlen( resourcedir ) );
	}

	if( flags & RES_OPTIONS ) {
		filename = concatstr( path, "options.adeona" );
		if( allowed_chars_only( filename, strlen(filename), FILENAME_ALLOWED_CHARS ) != TRUE ||
			parse_options_filename( filename ) == FAILURE ) {
			load_default_options();
		}
		
        if( filename != NULL ) { 
            free( filename );
		    filename = NULL;	
        }
	}

	if( flags & RES_GATEWAYS ) {
		filename = concatstr( path, "gateways.adeona" );

		if( allowed_chars_only( filename, strlen(filename), FILENAME_ALLOWED_CHARS ) == TRUE ) {
			if( gateways_load( filename ) == SUCCESS ) {
			    memcpy( opts.gateways_filename, filename, strlen( filename ) );
            }
            else {
                load_default_gwfile();
            }
		}
		
        if( filename != NULL )  {
            free( filename );
		    filename = NULL;
        }
	}

	if( flags & RES_LANDMARKS ) {
		filename = concatstr( path, "landmarks.adeona" );

		if( allowed_chars_only( filename, strlen(filename), FILENAME_ALLOWED_CHARS ) == TRUE ) {
			if( landmarks_load( filename ) == FAILURE ) {
				load_default_lmfile();
			}
		}
		
        if( filename != NULL )  {
            free( filename );
		    filename = NULL;
        }
	}

	if( (flags & RES_WEBCAM) && opts.use_webcamimg ) {
		if( strlen( path ) + strlen( "isightcapture.exe" ) > MAX_STROPTION_LEN ) {
            info( "options webcam program path greater than %d bytes", MAX_STROPTION_LEN  );
		}
		else {
			memset( opts.webcamprog_path, 0x00, MAX_STROPTION_LEN );
			memcpy( opts.webcamprog_path, path, strlen( path ) ); 
			strncat( opts.webcamprog_path, "isightcapture.exe", MAX_STROPTION_LEN - strlen( path ) );

			if( allowed_chars_only( opts.webcamprog_path, strlen(opts.webcamprog_path), FILENAME_ALLOWED_CHARS ) != TRUE ) {
				error( "webcam program path contains illegal characters %s", opts.webcamprog_path ); 
			}
		}
	}

	return SUCCESS;
}


int parse_options_filename( char* optionsFileName )
{
    FILE* optFile = NULL;

	if( optionsFileName == NULL ) {
	    error( "invalid input" );
        return FAILURE;
    }

    if( (optFile = fopen( optionsFileName, "r" )) == NULL ) {
        error( "could not open options file" );
        return FAILURE;
    }

    int rv = parse_options_file( optFile );
    fclose( optFile );
    return rv;
}

int seek_past_options( FILE* optFile )
{

	if( optFile == NULL ) {
	    error( "invalid input" );
        return FAILURE;
    }
    
	char* optline = NULL;
    int optlinelen = 0; 
#if defined( __APPLE__ ) 
    while( ( optline = fgetln( optFile, (size_t*)&optlinelen ) ) != NULL )
#else
    while( ( optlinelen = getline( &optline, (size_t*)&optlinelen, optFile ) ) != -1 )
#endif
    {
        if( memcmp( optline, OPTION_END, strlen( OPTION_END ) ) == 0 ) {
#if !defined( __APPLE__ ) 
            if( optline != NULL ) { 
                free( optline );
                optline = NULL;
            }
#endif
            return SUCCESS;
        }
    }

#if !defined( __APPLE__ ) 
    if( optline != NULL ) {
        free( optline );
        optline = NULL;
    }
#endif

    return FAILURE;
}

int parse_options_file( FILE* optFile )
{

    if( optFile == NULL ) {
	    error( "invalid input" );
        return FAILURE;
    }
    
    init_opts();

    char* optline = NULL;
    int optlinelen = 0; 
#if defined( __APPLE__ ) 
    while( ( optline = fgetln( optFile, (size_t*)&optlinelen ) ) != NULL )
#else
    while( ( optlinelen = getline( &optline, (size_t*)&optlinelen, optFile ) ) != -1 )
#endif
    {
        if( memcmp( optline, OPTION_END, strlen( OPTION_END ) ) == 0 ) {
            break;
        }
        else if( memcmp( optline, OPTION_CTXT_QUEUE_SIZE, strlen( OPTION_CTXT_QUEUE_SIZE ) ) == 0 )
        {
            //TODO: Allow CTXT_QUEUE_SIZE == 0
            if( sscanf( optline,  OPTION_CTXT_QUEUE_SIZE " = " INT_OPTION_FMT, &opts.ctxt_queue_size ) > 0 ) {
                if( opts.ctxt_queue_size <= 0 ) { 
                    info( "Invalid CTXT_QUEUE_SIZE value, using default" );
                    opts.ctxt_queue_size = DEFAULT_CTXT_QUEUE_SIZE;
                }
            }
            else {
                opts.ctxt_queue_size = DEFAULT_CTXT_QUEUE_SIZE;
            }
            
            info( "CTXT_QUEUE_SIZE = %d", opts.ctxt_queue_size );
        }
        else if( memcmp( optline, OPTION_SCHED_LAMBDA, strlen( OPTION_SCHED_LAMBDA ) ) == 0 )
        {
            if( sscanf( optline, OPTION_SCHED_LAMBDA " = " FLT_OPTION_FMT, &opts.sched_lambda ) > 0 ) {
                if( opts.sched_lambda < 0.0f ) {
                    info( "SCHED_LAMBDA set to less than 0, using default" );
                    opts.sched_lambda = DEFAULT_SCHED_LAMBDA;
                }
            }
            else {
                opts.sched_lambda = DEFAULT_SCHED_LAMBDA;
            }
            
            info( "SCHED_LAMBDA = %f", opts.sched_lambda  );
        }
        else if( memcmp( optline, OPTION_SERVER_PORT, strlen( OPTION_SERVER_PORT ) ) == 0 )
        {
            if( sscanf( optline, OPTION_SERVER_PORT " = " INT_OPTION_FMT, &opts.server_port ) > 0 ) {
                if( opts.server_port < 0 ) {
                    info( "Invalid port number, using default" );
                    opts.server_port = DEFAULT_SERVER_PORT;
                }
            }
            else {
                opts.server_port = DEFAULT_SERVER_PORT;
            }
            
            info( "SERVER_PORT = %d", opts.server_port );
        }
        else if( memcmp( optline, OPTION_SERVER_NAME, strlen( OPTION_SERVER_NAME ) ) == 0 )
        {
            if( optlinelen - strlen( OPTION_SERVER_NAME ) > MAX_STROPTION_LEN ) {
                info( "options specify server name greater than %d bytes", MAX_STROPTION_LEN  );
            }
            else if( sscanf( optline, OPTION_SERVER_NAME " = " STR_OPTION_FMT, opts.server_name ) > 0 ) {
                if( allowed_chars_only( opts.server_name, MAX_STROPTION_LEN, HOSTNAME_ALLOWED_CHARS ) == FALSE ) {
                    info( "Invalid server name, using default server" );
                    memcpy( opts.server_name, DEFAULT_SERVER_NAME, strlen( DEFAULT_SERVER_NAME ) + 1 ); //copy null terminator as well
                }
            }
            else {
                memcpy( opts.server_name, DEFAULT_SERVER_NAME, strlen( DEFAULT_SERVER_NAME ) + 1 ); //copy null terminator as well
            }

            info( "SERVER_NAME = %s", opts.server_name );
        }
        else if( memcmp( optline, OPTION_NUM_DHTKEYS_TOTRY, strlen( OPTION_NUM_DHTKEYS_TOTRY ) ) == 0 )
        {
            if( sscanf( optline, OPTION_NUM_DHTKEYS_TOTRY " = " INT_OPTION_FMT, &opts.num_dhtkeys_totry ) > 0 ) {
                if( opts.num_dhtkeys_totry < 0 ) {
                    info( "NUM_DHTKEYS_TOTRY is negative, using default" );
                    opts.num_dhtkeys_totry = DEFAULT_NUM_DHTKEYS_TOTRY;
                }
            }
            else {
                opts.num_dhtkeys_totry = DEFAULT_NUM_DHTKEYS_TOTRY;
            }
            
            info( "NUM_DHTKEYS_TOTRY = %d", opts.num_dhtkeys_totry );
        }
        else if( memcmp( optline, OPTION_USE_TRACEROUTES, strlen( OPTION_USE_TRACEROUTES ) ) == 0 )
        {
            if( sscanf( optline, OPTION_USE_TRACEROUTES " = " INT_OPTION_FMT, &opts.use_traceroutes ) > 0 ) {
                if( opts.use_traceroutes != 1 && opts.use_traceroutes != 0 ) {
                    opts.use_traceroutes = DEFAULT_USE_TRACEROUTES;
                }
            }
            else {
                opts.use_traceroutes = DEFAULT_USE_TRACEROUTES;
            }
            
            info( "USE_TRACEROUTES = %d", opts.use_traceroutes );
        }
//        else if( memcmp( optline, OPTION_LANDMARKS_FILENAME, strlen( OPTION_LANDMARKS_FILENAME ) ) == 0 )
//        {
//            if( optlinelen - strlen( OPTION_LANDMARKS_FILENAME ) > MAX_STROPTION_LEN ) {
//                info( "options specify landmarks filename greater than %d bytes", MAX_STROPTION_LEN  );
//            }
//            else if( sscanf( optline, OPTION_LANDMARKS_FILENAME " = " STR_OPTION_FMT, opts.landmarks_filename ) > 0 ) {
//                if( allowed_chars_only( opts.landmarks_filename, MAX_STROPTION_LEN, FILENAME_ALLOWED_CHARS ) == TRUE ) {
//                    if( landmarks_load( opts.landmarks_filename ) == SUCCESS ) {
//                        info( "LANDMARKS_FILENAME = %s", opts.landmarks_filename );
//                            continue;
//                    }
//                }
//            }
//            
//            info( "failed to load specified landmarks file" ); 
//            load_default_lmfile();
//        }
        else if( memcmp( optline, OPTION_MAX_SLEEPTIME, strlen( OPTION_MAX_SLEEPTIME ) ) == 0 )
        {
            if( sscanf( optline, OPTION_MAX_SLEEPTIME " = " INT_OPTION_FMT, &opts.max_sleeptime ) > 0 ) {
                // Disallow less than 300 seconds
                if( opts.max_sleeptime < MIN_SLEEPTIME ) { 
                    info( "MAX_SLEEPTIME too small, using default" );
                    opts.max_sleeptime = DEFAULT_MAX_SLEEPTIME;
                }
            }
            else {
                opts.max_sleeptime = DEFAULT_MAX_SLEEPTIME;
            }

            info( "MAX_SLEEPTIME = %d", opts.max_sleeptime );
        }
//        else if( memcmp( optline, OPTION_GATEWAYS_FILENAME, strlen( OPTION_GATEWAYS_FILENAME ) ) == 0 )
//        {
//            if( optlinelen - strlen( OPTION_LOG_FILENAME ) > MAX_STROPTION_LEN ) {
//                info( "options specify gateway filename greater than %d bytes", MAX_STROPTION_LEN  );
//            }
//            else if( sscanf( optline, OPTION_GATEWAYS_FILENAME " = " STR_OPTION_FMT, opts.gateways_filename ) > 0 ) {
//                if( allowed_chars_only( opts.gateways_filename, MAX_STROPTION_LEN, FILENAME_ALLOWED_CHARS ) == TRUE ) {
//                    if( gateways_load( opts.gateways_filename ) == SUCCESS ) {
//                        info( "GATEWAYS_FILENAME = %s", opts.gateways_filename );
//                        continue;
//                    }
//                }
//            }
//            info( "failed to load specified gateway file" );
//            load_default_gwfile();
//        }
        else if( memcmp( optline, OPTION_USE_WEBCAMIMG, strlen( OPTION_USE_WEBCAMIMG ) ) == 0 )
        {
            if( sscanf( optline, OPTION_USE_WEBCAMIMG " = " INT_OPTION_FMT, &opts.use_webcamimg ) > 0 ) {
                if( opts.use_webcamimg != 1 && opts.use_webcamimg != 0 ) {
                    opts.use_webcamimg = DEFAULT_USE_WEBCAMIMG;
                }
            }
            else {
                opts.use_webcamimg = DEFAULT_USE_WEBCAMIMG;
            }
            
            info( "USE_WEBCAMIMG = %d", opts.use_webcamimg );
        }
//        else if( memcmp( optline, OPTION_WEBCAMPROG_PATH, strlen( OPTION_WEBCAMPROG_PATH ) ) == 0 )
//        {
//            if( optlinelen - strlen( OPTION_WEBCAMPROG_PATH ) > MAX_STROPTION_LEN ) {
//                info( "options specify landmarks filename greater than %d bytes", MAX_STROPTION_LEN  );
//            }
//            else if( sscanf( optline, OPTION_WEBCAMPROG_PATH " = " STR_OPTION_FMT, opts.webcamprog_path ) > 0 ) {
//                if( allowed_chars_only( opts.webcamprog_path, MAX_STROPTION_LEN, FILENAME_ALLOWED_CHARS ) == TRUE ) {
//                    info( "WEBCAMPROG_PATH = %s", opts.webcamprog_path );
//                    continue;
//                }
//            }
//        }
        else if( memcmp( optline, OPTION_GWUPDATE_URL, strlen( OPTION_GWUPDATE_URL ) ) == 0 )
        {
            if( optlinelen - strlen( OPTION_GWUPDATE_URL ) > MAX_STROPTION_LEN ) {
                info( "options specify gwupdate url greater than %d bytes", MAX_STROPTION_LEN  );
            }
            else if( sscanf( optline, OPTION_GWUPDATE_URL " = " STR_OPTION_FMT, opts.gwupdate_url ) > 0 ) {
                if( allowed_chars_only( opts.gwupdate_url, MAX_STROPTION_LEN, HOSTNAME_ALLOWED_CHARS "/" ) == FALSE ) {
                    info( "Invalid url, using default " DEFAULT_GWUPDATE_URL );
                    memcpy( opts.gwupdate_url, DEFAULT_GWUPDATE_URL, strlen( DEFAULT_GWUPDATE_URL ) + 1 ); //copy null terminator as well
                }
            }
            else {
                memcpy( opts.gwupdate_url, DEFAULT_GWUPDATE_URL, strlen( DEFAULT_GWUPDATE_URL ) + 1 ); //copy null terminator as well
            }

            info( "GWUPDATE_URL = %s", opts.gwupdate_url );
        }
        else if( memcmp( optline, OPTION_USE_GWUPDATE, strlen( OPTION_USE_GWUPDATE ) ) == 0 )
        {
            if( sscanf( optline, OPTION_USE_GWUPDATE " = " INT_OPTION_FMT, &opts.use_gwupdate ) > 0 ) {
                if( opts.use_gwupdate != 1 && opts.use_gwupdate != 0 ) {
                    opts.use_gwupdate = DEFAULT_USE_GWUPDATE;
                }
            }
            else {
                opts.use_gwupdate = DEFAULT_USE_GWUPDATE;
            }
            
            info( "USE_GWUPDATE = %d", opts.use_gwupdate );
        }
        else if( memcmp( optline, OPTION_EXTIP_URL, strlen( OPTION_EXTIP_URL ) ) == 0 )
        {
            if( optlinelen - strlen( OPTION_EXTIP_URL ) > MAX_STROPTION_LEN ) {
                info( "options specify extip url greater than %d bytes", MAX_STROPTION_LEN  );
            }
            else if( sscanf( optline, OPTION_EXTIP_URL " = " STR_OPTION_FMT, opts.extip_url ) > 0 ) {
                if( allowed_chars_only( opts.extip_url, MAX_STROPTION_LEN, HOSTNAME_ALLOWED_CHARS "/" ) == FALSE ) {
                    info( "Invalid url, using default " DEFAULT_EXTIP_URL );
                    memcpy( opts.extip_url, DEFAULT_EXTIP_URL, strlen( DEFAULT_EXTIP_URL ) + 1 ); //copy null terminator as well
                }
            }
            else {
                memcpy( opts.extip_url, DEFAULT_EXTIP_URL, strlen( DEFAULT_EXTIP_URL ) + 1 ); //copy null terminator as well
            }

            info( "EXTIP_URL = %s", opts.extip_url );
        }
        else {
            error( "unrecognized option" );
            return FAILURE;
        }
        
        memset( optline, 0x00, optlinelen );
    }
    

#if !defined( __APPLE__ ) 
    if( optline != NULL )  { 
        free( optline );
        optline = NULL;
    }
#endif

    return SUCCESS;
}

