/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include "common.h"
#include "config.h"
#include "util.h"
#include "log.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <openssl/rand.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <unistd.h>
#if defined( __CYGWIN__ )
    #include <windows.h>
#endif

// big endian from apple big endian macro or config.h
#if ( defined( __APPLE__ ) && defined( __BIG_ENDIAN__ ) ) || defined( WORDS_BIGENDIAN )
    #define IS_BIGENDIAN 1
#else
    #define IS_BIGENDIAN 0
#endif

#define MAX_PASSWDASK 6

static void reverse_bytes( uchar* left, int len )
{
    assert( left != NULL && len > 0 );
    
    uchar* right = left + (len - 1);
    while( left < right ) {
        uchar tmp = *left;
        *left = *right;
        *right = tmp;
        right--;
        left++;
    }
}

int hex2bin( uchar* hex IN, int hlen IN, uchar* bin OUT, int blen IN )
{
    if( hex != NULL && bin != NULL && hlen > 0 && blen > 0 ) {
        if( hlen%2 == 0 && hlen%blen == 0 ) {
            int i;
            for( i=0; i<blen; i++ ) {
                unsigned int tmp;
                if( sscanf( (char*)hex, "%02X", &tmp ) != 0 ) {
                    assert( tmp <= 255 );
                    bin[i] = (uchar)tmp;
                    hex += 2;
                }
                else {
                    error( "could not match hex char" );       
                    return FAILURE;
                }
            }
        }
        else {
            error( "invalid length parameter" );
            return FAILURE;
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

int bin2hex( uchar* hex OUT, int hlen IN, uchar* bin IN, int blen IN )
{
    if( hex != NULL && bin != NULL && hlen > 0 && blen > 0 ) {
        if( hlen%2 == 0 && hlen%blen == 0 ) {
            int i;
            for( i=0; i<blen; i++ ) {
                snprintf( (char*)hex, 3, "%02X", bin[i] );
                hex += 2;
            }
        } 
        else {
            error( "invalid length parameter" );
            return FAILURE;
        }

        return SUCCESS;
    }
    else {
        error( " invalid input" );
        return FAILURE;
    }
}

int fprintln_bin2hex( FILE* file IN, uchar* bin IN, int len IN )
{
    if( file != NULL && bin != NULL && len > 0 ) {
        int i;
        for( i=0; i<len; i++ ) {
            fprintf( file, "%02X", bin[i] );
        }
        fprintf( file, NEWLINE_CHAR );
        
        return SUCCESS;
    }
    else {
		error( "invalid input" );
        return FAILURE;
    }
}

/*
 * This will serialize <data> in little endian byte order and write it
 * out as a hex string to <file>.
 *
 */
int fprintln_data2hex( FILE* file IN, void* data IN, int len IN )
{
    if( file != NULL && data != NULL && len > 0 ) {
        if( IS_BIGENDIAN ) {
            int i;
            for( i=( len-1 ); i>=0; i-- ) {
                fprintf( file, "%02X", ((uchar*)data)[i] );
            }
            fprintf( file, NEWLINE_CHAR );
        }
        else {
            if( fprintln_bin2hex( file, (uchar*)data, len ) == FAILURE ) {
                return FAILURE;
            }
        }

        return SUCCESS;
    }
    else {
		error( "invalid input" );
        return SUCCESS;
    }

}

int freadln_hex2bin( FILE* file IN, uchar* bin OUT, int len IN )
{
    if( file != NULL && bin != NULL && len > 0 ) {
        int expectedhexlen = len * 2;
        char* hexline = NULL;
        int hexlinelen = 0;

#if defined( __APPLE__ ) 
        hexline = fgetln( file, (size_t*)&hexlinelen );
        if( hexline == NULL ) {
            return FAILURE;
        }
#else
         hexlinelen = getline( &hexline, (size_t*)&hexlinelen, file );
         if( hexlinelen < 0 ) {
            return FAILURE;
         }
#endif
        hexline[expectedhexlen] = '\0'; // null terminate the hex string for sanity

        if( allowed_chars_only( hexline, expectedhexlen, "0123456789ABCDEF" ) == FALSE ) {
            error( "invalid hex chars read" );
            return FAILURE;
        }
        
        if( hex2bin( (uchar*)hexline, expectedhexlen, bin, len ) == FAILURE ) {
#if !defined( __APPLE__ )
            if( hexline != NULL ) {
                free( hexline );
                hexline = NULL;
            }
#endif
            return FAILURE;
        }

        return SUCCESS;
    }
    else {
		error( "invalid input" );
        return FAILURE;
    }
}

/*
 * This will deserialize a hex string (representing a data type i.e.
 * int) in little endian byte order, to the platform
 * endianness and write it out to <data>.
 *
 */
int freadln_hex2data( FILE* file IN, void* data OUT, int len IN )
{
    if( file != NULL && data != NULL && len > 0 ) {
        
        if( freadln_hex2bin( file, (uchar*)data, len ) == FAILURE ) {
            return FAILURE;
        }
        
        if( IS_BIGENDIAN ) {
            reverse_bytes( (uchar*)data, len );
        }

        return SUCCESS;
    }
    else {
		error( "invalid input" );
        return FAILURE;
    }
}

/*
 * memcpy but accounts for endianness.
 * 
 */
void datacpy( void* dst IN, void* src IN, int len IN )
{
    if( dst != NULL && src != NULL && len > 0 ) {
        memcpy( dst, src, len );
        
        if( IS_BIGENDIAN ) {
            reverse_bytes( (uchar*)dst, len );
        }
    }
    else {
		error( "invalid input" );
    }

}

/*
 * Use the openssl rand library to generate "cryptographically strong"
 * pseudo-random bytes. 
 * If openssl fails, <buf> is cleared.
 */ 
int get_random_bytes( uchar* buf OUT, int len IN )
{
    if( buf != NULL && len > 0 ) {
        if( !RAND_bytes( buf, len ) ) {
            memset( buf, 0x00, len );
            return FAILURE;
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Random number in [0, max)
 *
 */ 
unsigned int get_random_uint( unsigned int max )
{
    unsigned int r;
    get_random_bytes( (uchar*)&r, sizeof( unsigned int ) );
    
    return ( r % max );
}

/*
 * Note the static buffer for output.
 *
 */ 
char* time2string( time_t time IN, int utc )
{
    static char timestr[128];
    memset( timestr, 0x00, sizeof( timestr ) );
    
    struct tm* taym = ( utc == TRUE ) ?  gmtime( &time ) : localtime( &time );
    strftime( timestr, sizeof( timestr ), "%m/%d/%Y,%H:%M (%Z)", taym );
    
    return timestr;
}

/*
 * String output suitable for filenames 
 * Note the static buffer for output.
 *
 */ 
char* time2fstring( time_t time IN, int utc )
{
    static char timestr[128];
    memset( timestr, 0x00, sizeof( timestr ) );
    
    struct tm* taym = ( utc == TRUE ) ?  gmtime( &time ) : localtime( &time );
    strftime( timestr, sizeof( timestr ), "%m-%d-%Y-%H%M-%Z", taym );
    
    return timestr;
}

/*
 * Parses <timestr> expecting the following format: 
 * MM/DD/YYYY,HH:MM
 *
 */ 
time_t string2time( char* timestr IN )
{
    if( timestr != NULL ) {
        struct tm brokedown;
        memset( &brokedown, 0x00, sizeof( struct tm ) );
        
        //strptime( timestr, "%m/%d/%Y,%H:%M", &brokedown );        
        sscanf( timestr, "%d/%d/%d,%d:%d", 
                &brokedown.tm_mon, &brokedown.tm_mday, &brokedown.tm_year, 
                &brokedown.tm_hour, &brokedown.tm_min );
        
        brokedown.tm_mon  -= 1; 
        brokedown.tm_year -= 1900;

        if( brokedown.tm_mon < 0 || brokedown.tm_mon > 11 
            || brokedown.tm_mday < 1 || brokedown.tm_mday > 31 
            || brokedown.tm_year < 0 
            || brokedown.tm_hour < 0 || brokedown.tm_hour > 23 
            || brokedown.tm_min < 0 || brokedown.tm_min > 59 )
        {
            error( "invalid time string" );
            return (time_t)-1;
        }
        
        //check for DST
        time_t now = time( NULL );
        brokedown.tm_isdst = localtime( &now )->tm_isdst;
        
        return mktime( &brokedown );
    }
    else {
        error( "invalid input" );
        return (time_t)-1;
    }
}

/*
 * Fork a process, execute a command and write the output to <buffer>
 * 
 * WARNING: <cmd> is UNCHECKED and passed directly to popen 
 *
 * NOTE: The Windows code does not set hStdInput of STARTUPINFOA but
 * it should be okay since pathping.exe doesn't expect stdin
 *
 */ 
int exec_command( char* cmd IN, char* buffer OUT, int blen OUT )
{

#if defined( __CYGWIN__ )

    if( cmd != NULL && buffer != NULL && blen > 0 ) {
        // Reference: http://support.microsoft.com/default.aspx?scid=kb;en-us;Q190351
        // TODO: Should distinguish #define of TRUE/FALSE, works for now though

        // first, setup pipe
        HANDLE hOutputReadTmp, hOutputRead, hOutputWrite;
        HANDLE hErrorWrite;
        SECURITY_ATTRIBUTES sa;

        sa.nLength= sizeof( SECURITY_ATTRIBUTES );
        sa.lpSecurityDescriptor = NULL;
        sa.bInheritHandle = TRUE; 
        if ( !CreatePipe( &hOutputReadTmp, &hOutputWrite, &sa, 0 ) ) {
            error( "could not create pipe" );
            return FAILURE;
        }
        
        // reference says this is necessary in case child closes
        // stdout handles, pretty clever
        if( !DuplicateHandle( GetCurrentProcess(), hOutputWrite,
                              GetCurrentProcess(), &hErrorWrite, 
                              0, TRUE, DUPLICATE_SAME_ACCESS ) )
        {
            error( "could not duplicate handle" );
            CloseHandle( hOutputReadTmp );
            CloseHandle( hOutputWrite );
            return FAILURE;
        }
        
        // we don't want the child to inherit the read end of the pipe
        // so we create a non-inheritable copy
        if( !DuplicateHandle( GetCurrentProcess(), hOutputReadTmp,
                              GetCurrentProcess(), &hOutputRead, 
                              0, FALSE, DUPLICATE_SAME_ACCESS ) )
        {
            error( "could not duplicate handle" );
            CloseHandle( hOutputReadTmp );
            CloseHandle( hOutputWrite );
            CloseHandle( hErrorWrite );
            return FAILURE;
        }
        
        // now close the inheritable copy of the read end of the pipe
        if( !CloseHandle( hOutputReadTmp ) ) {
            error( "could not close handle" );
            CloseHandle( hOutputWrite );
            CloseHandle( hErrorWrite );
            CloseHandle( hOutputRead );
            return FAILURE;
        }
    
        //now, create the child process
        PROCESS_INFORMATION pi;
        STARTUPINFOA si;
        
        ZeroMemory( &si, sizeof( STARTUPINFO ) );
        si.cb = sizeof( STARTUPINFO );
        si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
        si.hStdOutput = hOutputWrite;
        si.hStdError = hErrorWrite;
        si.wShowWindow = SW_HIDE;
    
        if( !CreateProcessA( NULL, cmd, NULL, NULL, TRUE, CREATE_NEW_CONSOLE, 
                             NULL, NULL , &si, &pi ) )
        {
            error( "could not create process" );
            CloseHandle( hOutputWrite );
            CloseHandle( hErrorWrite );
            CloseHandle( hOutputRead );
            return FAILURE;
        }
        
        //unnecessary handle
        if( !CloseHandle( pi.hThread ) ) {
            error( "could not close handle" );
        }
        
        // reference says ReadFile will hang if we maintain write
        // handles in the parent because it will cause the pipe to not close
        if( !CloseHandle( hOutputWrite ) || !CloseHandle( hErrorWrite ) ) {
            error( "could not close handles" );
            CloseHandle( hOutputRead );
            CloseHandle( pi.hProcess );
            return FAILURE;
        }
        
        //finally, read the child's output
        DWORD nBytesRead = 0;
        char* bread = buffer;
        while( ReadFile( hOutputRead, bread, blen, &nBytesRead, NULL ) > 0 ) {
            bread += nBytesRead;
        }
        
        if( GetLastError() != ERROR_BROKEN_PIPE ) {
            error( "could not complete read from pipe" );
            memset( buffer, 0x00, blen );
            CloseHandle( hOutputRead );
            CloseHandle( pi.hProcess );
            return FAILURE;
        }

        if( WaitForSingleObject( pi.hProcess, INFINITE ) == WAIT_FAILED ) {
            error( "could not wait for child process" );
            CloseHandle( hOutputRead );
            CloseHandle( pi.hProcess );
            return FAILURE;
        }
        
        if( !CloseHandle( hOutputRead ) ) {
            error( "could not close handle" );
            CloseHandle( pi.hProcess );
            return FAILURE;
        }
        
        if( !CloseHandle( pi.hProcess ) ) {
            error( "could not close handle" );
            return FAILURE;
        }
        
        return SUCCESS;
    }

#else 
    
    if( cmd != NULL && buffer != NULL && blen > 0 ) {
        fflush( NULL ); //recommended by man page
        
        FILE* fout = popen( cmd, "r" );
        if(  fout == NULL ) {
            error( "cannot execute command" );
            return FAILURE;
        }
        
        int nread = 0;
        char* bread = buffer;
        int nbread = blen;
        while( ( nread  = read( fileno( fout ), bread, nbread ) ) > 0 ) {
            if( nread < 0 ) {
                error( "cannot read from pipe" );
                memset( buffer, 0x00, blen );
                return FAILURE;
            }

            nbread -= nread;
            assert( nbread >= 0 );
            bread += nread;
        }
        
        int child = pclose( fout );
        if( !WIFEXITED( child ) ) {
            error( "\"%s\" exited abnormally", cmd );
            memset( buffer, 0x00, blen );
            return FAILURE;
        }
        
        return SUCCESS;
    }

#endif
    
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * File position indicator is moved back to 0 on return.
 *
 */
int count_lines_in_file( FILE* file IN )
{
    int c = 0;
    int count = 0;
    
    while( ( c = fgetc( file ) ) != EOF ) {
        if( (char)c == '\n' ) {
            count++;
        }
    }

    rewind( file );

    return count;
}

/*
 * Returns a new null-terminated string that is (str1||str2), caller
 * must deallocate
 *
 */ 
char* concatstr( char* str1 IN, char* str2 IN )
{
    if( str1 != NULL && str2 != NULL ) {
        int len1 = strlen( str1 );
        int len2 = strlen( str2 );
        
        char* newstr = (char*)malloc( len1 + len2 + 1 );
        assert( newstr != NULL );

        memset( newstr, 0x00, len1 + len2 + 1 );
        memcpy( newstr, str1, len1 );
        memcpy( &newstr[len1], str2, len2 );
        
        return newstr;
    }
    else {
        error( "invalid input" );
        return NULL;
    }
}

/*
 * Prompt user for password string. 
 * <pwdbuf> must be allocated by caller.
 * <plen> IN is the length of <pwdbuf> and <plen> OUT is the length of
 * the password string the user gave.
 * 
 */
int pwdprompt_loop( char* pwdbuf OUT, int* plen INOUT, int verify IN )
{
    if( pwdbuf != NULL && plen != NULL && *plen > 0 ) {
        memset( pwdbuf, 0x00, *plen );        
        
        int ntry = 0;
        while( ntry++ < MAX_PASSWDASK && EVP_read_pw_string( pwdbuf, *plen, "Please enter password for Adeona: ", verify ) != 0 ) {
            const char *data;
            int flags;
            if( ERR_get_error_line_data( NULL, NULL, &data, &flags ) != 0 ) {
                if( data != NULL && ( flags & ERR_TXT_STRING ) ) {
                    printout( "%s\n", data );
                }
                else {
                    printout( "Sorry, invalid input\n\n" );
                }
            }
        }
        
        int pslen = strlen( pwdbuf );
        assert( pslen >= 0 && pslen <= *plen );
        
        if( pslen == 0 || ntry >= MAX_PASSWDASK ) {
            printout( "\nSorry, could not get a valid password.\n" );
            memset( pwdbuf, 0x00, *plen ); //no password given, clear buffer
            pslen = 0;
        }

        *plen = pslen;

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Flips msb of an unsigned integer
 *
 */
uint flipmsb( uint x ) 
{
    return x = x ^ 0x80000000;
}

/* 
 * Checks MSB
 *
 */
int ismsbset( uint x ) 
{
    if( x & 0x80000000 ) 
        return SUCCESS;
    return FAILURE;
}

/*
 * strspn wrapper
 * <string> and <allowed> should be NULL terminated
 * 
 * <slen> is in case <string> is not NULL terminated
 *
 */ 
int allowed_chars_only( const char* string IN, int slen, const char* allowed IN )
{
    if( string != NULL && allowed != NULL ) {
        int numgood = strspn( string, allowed );

        if( numgood != strlen( string ) || numgood > slen ) {
            return FALSE;
        }
        
        return TRUE;
    }
    else {
        error( "invalid input" );
        return FALSE;
    }
}

