/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */

#include "common.h"
#include "state.h"
#include "util.h"
#include "log.h"
#include "ctxtcache.h"
#include <math.h>
#include "sched.h"

static void print_state_contents( state_t* state )
{
    debug( "update time: %s", time2string( state->updatetime, FALSE ) );
    debug( "update time: %d", (int)state->updatetime );
    log_msg_bin( ADEONA_DEBUG, state->currentseed, sizeof( prgkey_t ),   "currentseed: " );
    log_msg_bin( ADEONA_DEBUG, state->cacheseed, sizeof( prgkey_t ),     "cacheseed:   " );
    
    if( state->sd != NULL ) {
        log_msg_bin( ADEONA_DEBUG, (uchar*)state->sd->nextseed, sizeof( prgkey_t ), "nextseed: " );
        log_msg_bin( ADEONA_DEBUG, (uchar*)state->sd->nextcacheseed, sizeof( prgkey_t ), "nextcacheseed: " );
        log_msg_bin( ADEONA_DEBUG, (uchar*)state->sd->schedseed, sizeof( prgblock_t ), "schedseed: " );
        log_msg_bin( ADEONA_DEBUG, (uchar*)state->sd->renckey, sizeof( renckey_t ), "reenckey: " );
        log_msg_bin( ADEONA_DEBUG, (uchar*)state->sd->dhtkeys, sizeof( dhtkey_t ) * opts.num_dhtkeys_totry, "dhtkeys: " );
    }
    
    if( state->cache != NULL ) {
        log_msg_bin( ADEONA_DEBUG, (uchar*)state->cache, CTXT_QUEUE_LEN, "cache:       " );
    }
    
}

//#define STATE_FILENAME "statefile_test_from_x86.txt"
#define STATE_FILENAME "statefile_test_from_powerpc.txt"
static void make_statefile( void )
{   
    state_t* state = state_alloc();
    assert( state != NULL );
    
    state_makenew( state );
    
    print_state_contents( state );
    
    state_save_filename( STATE_FILENAME, state, NULL, 0, SC_ALL );
    
    state_free( state );
}

static void load_statefile( void )
{
    state_t* state = state_alloc();
    
    state_load_filename( STATE_FILENAME, state, NULL, 0, SC_ALL );

    print_state_contents( state );
    
    state_free( state );
}

static void test_saveload( void )
{
    state_t* state = state_alloc();
    assert( state != NULL );

    assert( state_makenew( state ) == SUCCESS );
    
    state_save_filename( "statefile_nopass", state, NULL, 0, SC_ALL );
    state_save_filename( "statefile_yespass", state, "password", strlen( "password" ), SC_ALL );
    
    char* password = "password";
    state_t* nopass = state_alloc();
    state_t* yespass = state_alloc();
    state_t* badpass = state_alloc();
    state_t* nocache = state_alloc();
    assert( nopass != NULL &&  yespass != NULL && badpass != NULL && nocache != NULL );
     
    assert( state_load_filename( "statefile_nopass", nopass, NULL, 0, SC_ALL ) == SUCCESS );
    assert( state_load_filename( "statefile_yespass", yespass, password, strlen( password ), SC_ALL ) == SUCCESS );
    assert( state_load_filename( "statefile_yespass", badpass, "badpassword", strlen( "badpassword" ), SC_ALL ) == FAILURE );
    assert( state_load_filename( "statefile_yespass", nocache, password, strlen( password ), SC_NOCACHE ) == SUCCESS );
    
    assert( nocache->cache != NULL ); //because of state_alloc, maybe it should get flags too? might get too confusing
    
    assert( state->updatetime == nopass->updatetime
            && state->updatetime == yespass->updatetime
            && state->updatetime != badpass->updatetime
            && state->updatetime == nocache->updatetime );
    
    assert( memcmp( state->currentseed, nopass->currentseed, sizeof( prgkey_t ) ) == 0 
            && memcmp( state->currentseed, yespass->currentseed, sizeof( prgkey_t ) ) == 0 
            && memcmp( state->currentseed, badpass->currentseed, sizeof( prgkey_t ) ) != 0
            && memcmp( state->currentseed, nocache->currentseed, sizeof( prgkey_t ) ) == 0 );
    
    assert( memcmp( state->cacheseed, nopass->cacheseed, sizeof( prgkey_t ) ) == 0 
            && memcmp( state->cacheseed, yespass->cacheseed, sizeof( prgkey_t ) ) == 0 
            && memcmp( state->cacheseed, badpass->cacheseed, sizeof( prgkey_t ) ) != 0
            && memcmp( state->cacheseed, nocache->cacheseed, sizeof( prgkey_t ) ) == 0 );

    assert( memcmp( state->cache, nopass->cache, CTXT_QUEUE_LEN ) == 0 
            && memcmp( state->cache, yespass->cache, CTXT_QUEUE_LEN ) == 0 
            && memcmp( state->cache, badpass->cache, CTXT_QUEUE_LEN ) != 0 
            && memcmp( state->cache, nocache->cache, CTXT_QUEUE_LEN ) != 0 );
    
    state_free( state );
    state_free( nopass  );
    state_free( yespass );
    state_free( badpass );
    state_free( nocache );
    
    system( "rm statefile_nopass" );
    system( "rm statefile_yespass" );

    printout( "***%s tests passed\n", __FUNCTION__ );
}

static void test_stateseeds( void )
{
    //same test vectors in crypto module
    prgkey_t test1_1 = "\x66\xE9\x4B\xD4\xEF\x8A\x2C\x3B\x88\x4C\xFA\x59\xCA\x34\x2B\x2E"; 
    prgkey_t test1_2 = "\x47\x71\x18\x16\xE9\x1D\x6F\xF0\x59\xBB\xBF\x2B\xF5\x8E\x0F\xD3";
    prgkey_t test2_1 = "\xAB\x4B\x36\xBD\xC8\x1D\x8B\xD2\xBC\xED\x4B\x05\xA7\x08\xB4\x22";
    prgkey_t test2_2 = "\xC2\x16\x7E\x7A\xE6\xF0\xF3\x66\xEC\x26\x45\x07\x75\x78\x3D\xDA";
    
    state_t* state = state_alloc();
    assert( state != NULL );
    
    //state->sd->nextseed is zeroed out by state_alloc
    sched_nextstate( state );
    assert( memcmp( state->sd->nextseed, test1_1, sizeof( prgkey_t ) ) == 0 
            && memcmp( state->sd->nextcacheseed, test1_2, sizeof( prgkey_t ) ) == 0 );
    
    //now nextseed should be test1_1 and we expect test2_*
    sched_nextstate( state );
    assert( memcmp( state->sd->nextseed, test2_1, sizeof( prgkey_t ) ) == 0 
            && memcmp( state->sd->nextcacheseed, test2_2, sizeof( prgkey_t ) ) == 0 );
    
    state_free( state );
    
    printout( "***%s tests passed\n", __FUNCTION__ );
}

static void test_cacheseeds( void )
{
    //same test vectors in crypto module
    prgkey_t test1_1 = "\x66\xE9\x4B\xD4\xEF\x8A\x2C\x3B\x88\x4C\xFA\x59\xCA\x34\x2B\x2E"; 
    prgkey_t test2_1 = "\xAB\x4B\x36\xBD\xC8\x1D\x8B\xD2\xBC\xED\x4B\x05\xA7\x08\xB4\x22";
    
    state_t* state = state_alloc();
    assert( state != NULL );
    update_t dummy;
    memset( &dummy, 0x00, sizeof( update_t ) );

    //state->sd->nextcacheseed is zeroed out by state_alloc
    ctxt_cache_update( state->sd->nextcacheseed, state->cache, &dummy ); 
    assert( memcmp( state->sd->nextcacheseed, test1_1, sizeof( prgkey_t ) ) == 0 );
    
    ctxt_cache_update( state->sd->nextcacheseed, state->cache, &dummy ); 
    assert( memcmp( state->sd->nextcacheseed, test2_1, sizeof( prgkey_t ) ) == 0 );
    
    state_free( state );

    printout( "***%s tests passed\n", __FUNCTION__ );
}

static void test_nextprevbacknstate( void )
{
    state_t* state1 = state_alloc(); //use as orig
    //need to call prg_generate because makenew does not do it for the init program
    state_makenew( state1 );
    assert( prg_generate( state1->currentseed, STATE_NEXTSEED_LOW, STATE_RENC_HIGH, (uchar*)state1->sd ) == SUCCESS );
    assert( prg_generate( state1->currentseed, STATE_DHT_LOW, STATE_DHT_HIGH, (uchar*)state1->sd->dhtkeys ) == SUCCESS );

    state_t* state2 = state_alloc();
    state_t* state3 = state_alloc();
    assert( state1 != NULL && state2 != NULL && state3 != NULL );
    
    state_copy( state2, state1, SC_ALL );
    state_copy( state3, state1, SC_ALL );
    
    int i;
    for( i=0; i<10; i++ ) {
        sched_nextstate( state2 );
        sched_nextstate( state3 );
    }

    //extra step for state3
    sched_nextstate( state3 );
    
    assert( state2->updatetime != state3->updatetime );
    assert( memcmp( state2->currentseed, state3->currentseed, sizeof( prgkey_t ) ) != 0 );
    assert( memcmp( state2->cacheseed, state3->cacheseed, sizeof( prgkey_t ) ) != 0 );
    assert( memcmp( state2->sd->nextseed, state3->sd->nextseed, sizeof( prgkey_t ) ) != 0 );
    assert( memcmp( state2->sd->nextcacheseed, state3->sd->nextcacheseed, sizeof( prgkey_t ) ) != 0 );
    assert( memcmp( state2->sd->schedseed, state3->sd->schedseed, sizeof( prgblock_t ) ) != 0 );
    assert( memcmp( state2->sd->renckey, state3->sd->renckey, sizeof( renckey_t ) ) != 0 );
    assert( memcmp( state2->sd->dhtkeys, state3->sd->dhtkeys, sizeof( dhtkey_t ) * opts.num_dhtkeys_totry ) != 0 );
    
    //now it should be the same as state2
    sched_prevstate( state1, state3 );

    assert( state2->updatetime == state3->updatetime );
    assert( memcmp( state2->currentseed, state3->currentseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state2->cacheseed, state3->cacheseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state2->sd->nextseed, state3->sd->nextseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state2->sd->nextcacheseed, state3->sd->nextcacheseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state2->sd->schedseed, state3->sd->schedseed, sizeof( prgblock_t ) ) == 0 );
    assert( memcmp( state2->sd->renckey, state3->sd->renckey, sizeof( renckey_t ) ) == 0 );
    assert( memcmp( state2->sd->dhtkeys, state3->sd->dhtkeys, sizeof( dhtkey_t ) * opts.num_dhtkeys_totry ) == 0 );
    
    //now we should be back to the original state
    sched_gobackn( state1, state3, 10 );
    
    assert( state1->updatetime == state3->updatetime );
    assert( memcmp( state1->currentseed, state3->currentseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state1->cacheseed, state3->cacheseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state1->sd->nextseed, state3->sd->nextseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state1->sd->nextcacheseed, state3->sd->nextcacheseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state1->sd->schedseed, state3->sd->schedseed, sizeof( prgblock_t ) ) == 0 );
    assert( memcmp( state1->sd->renckey, state3->sd->renckey, sizeof( renckey_t ) ) == 0 );
    assert( memcmp( state1->sd->dhtkeys, state3->sd->dhtkeys, sizeof( dhtkey_t ) * opts.num_dhtkeys_totry ) == 0 );
    
    //what if n is "too big"
    assert( sched_gobackn( state1, state2, 100 ) == FAILURE );
    
    assert( state1->updatetime == state2->updatetime );
    assert( memcmp( state1->currentseed, state2->currentseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state1->cacheseed, state2->cacheseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state1->sd->nextseed, state2->sd->nextseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state1->sd->nextcacheseed, state2->sd->nextcacheseed, sizeof( prgkey_t ) ) == 0 );
    assert( memcmp( state1->sd->schedseed, state2->sd->schedseed, sizeof( prgblock_t ) ) == 0 );
    assert( memcmp( state1->sd->renckey, state2->sd->renckey, sizeof( renckey_t ) ) == 0 );
    assert( memcmp( state1->sd->dhtkeys, state2->sd->dhtkeys, sizeof( dhtkey_t ) * opts.num_dhtkeys_totry ) == 0 );
    
    printout( "***%s tests passed\n", __FUNCTION__ );

}

int main( void )
{
    load_default_options();
//    make_statefile();
//    load_statefile();
    test_saveload();
    test_stateseeds();
    test_cacheseeds();
    test_nextprevbacknstate();
    
    exit( EXIT_SUCCESS );
}
