/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include "ctxtcache.h"
#include "log.h"
#include "util.h"
#include "location.h"

/*
 * Initialization of a ctxt_cache_t
 *
 * Allocates buffer for cache
 *
 * Fills the cache with random bits (this makes it indistinguishable
 * from an in-use cache)
 * 
 */
cache_elem_t* ctxt_cache_alloc( void )
{
    cache_elem_t* cache = (cache_elem_t*)malloc( CTXT_QUEUE_LEN ); 
    assert( cache != NULL );
    memset( cache, 0x00, CTXT_QUEUE_LEN );
    
    return cache;
}

/*
 * Fill ctxt cache with random bytes
 *
 */
int ctxt_cache_fillrand( cache_elem_t* cache OUT )
{
	if( cache != NULL  ) {
		prgkey_t tempkey;

		// Just use the PRG to fill the cache with pseudorandom bytes.
		// We could directly generate all the bytes randomly, but seems
		// like overkill.
		if( get_random_bytes( tempkey, sizeof( prgkey_t ) ) == FAILURE ) {
			error( "could not initialize cache --- no randomness generated " );
			return FAILURE;
		}

		if( prg_generate( tempkey, 0, CTXT_QUEUE_LEN - 1, (uchar*)cache ) == FAILURE ) {
			error( "prg_generate failed" );
			return FAILURE;
		}

		return SUCCESS;
	}
	else {
        error( "invalid input" );
        return FAILURE;
	}
}

/*
 * Write <cache> contents to <file>
 *
 */
int ctxt_cache_save( FILE* file IN, cache_elem_t* cache IN )
{
	if( file != NULL && cache != NULL ) {   
        fprintln_bin2hex( file, (uchar*)cache, CTXT_QUEUE_LEN );
        
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * 
 *
 */
int ctxt_cache_load( FILE* file IN, cache_elem_t* cache OUT )
{
	if( file != NULL && cache != NULL ) {
        freadln_hex2bin( file, (uchar*)cache, CTXT_QUEUE_LEN );
        
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Deep del of a ctxt_cache_t structure
 *
 */
int ctxt_cache_del( cache_elem_t* cache IN )
{
	if( cache != NULL ) {
		memset( cache, 0x00, CTXT_QUEUE_LEN );
        free( cache );
		cache = NULL;
	}

	return SUCCESS;
}

/*
 * Cache update
 *
 * Insert a new ctxt into cache
 * New ctxt is always of length CTXT_SIZE 
 *
 */
int ctxt_cache_update( prgkey_t cacheseed INOUT, cache_elem_t* cache INOUT, update_t* update IN )
{
	if( cacheseed != NULL && cache != NULL && update != NULL ) 
    {
		// Generate cache keys, sentinel value
		cache_keys_t keys;
		cache_elem_t elem;

        debug( "================== start cache update ======================" );       
		if( prg_generate( cacheseed, 0, sizeof( cache_keys_t ) - 1, (uchar*)&keys ) == FAILURE ) {
			error( "could not generate keys from cache seed" );
			return FAILURE;
		}
		memcpy( elem.sentinel, keys.sentinel, SENTINEL_SIZE );
	
		// Pad data appropriately to get plaintext
		uchar ptxt[PTXT_SIZE];
        memset( ptxt, 0x00, PTXT_SIZE );
		
        datacpy( ptxt, &update->skipped, sizeof( int ) );
        if( location_serialize( &update->loc, ptxt + sizeof( int ), LOC_DATA_UPPERBOUND ) == FAILURE ) {
            error( "serializing location failed" );
            memset( ptxt, 0x00, PTXT_SIZE );
        }
        
		// Encrypt data
		if( encrypt_then_mac( keys.enckey, 
							  keys.hmackey, 
							  ptxt, PTXT_SIZE, 
							  elem.ctxt, CTXT_SIZE ) == FAILURE ) 
        {
			error( "encrypt failed" );
//			return FAILURE;
		}
        
        log_msg_bin( ADEONA_DEBUG, (uchar*)&elem, sizeof( cache_elem_t ),     "new elem->cache[0] before reencryption:" );
        
        
		// Shift ctxt queue by 1 and reencrypt (if necessary)
		if( opts.ctxt_queue_size > 1 ) {
            log_msg_bin( ADEONA_DEBUG, (uchar*)&cache[0], sizeof( cache_elem_t ), "cache[0]->cache[1] before reencryption:" );
            
			memmove( &cache[1], &cache[0], CTXT_QUEUE_LEN - sizeof(cache_elem_t) );
			// Re-encrypt old part of cache 
			// this is just ctr-mode encryption with fixed IV
			// encryption function supports in-place encryption of a buffer
			//if( encryption( keys.renckey,
            //                (uchar*)&cache[1],
            //                (uchar*)&cache[1], 
            //                CTXT_QUEUE_LEN - sizeof(cache_elem_t) ) == FAILURE ) 
            //{
			//	error( "could not reencrypt" );
			//	return FAILURE;
			//}
		}
		
		// Place new ctxt in end of queue
		cache[0] = elem;
        
        if( opts.ctxt_queue_size > 1 ) {
            log_msg_bin( ADEONA_CRYPTO, cacheseed, sizeof( prgkey_t ),             "cacheseed :" );
            log_msg_bin( ADEONA_CRYPTO, keys.enckey, sizeof( enckey_t ),           "enckey    :" );
            log_msg_bin( ADEONA_CRYPTO, keys.hmackey, sizeof( hmackey_t ),         "hmackey   :" );
            //log_msg_bin( ADEONA_CRYPTO, keys.renckey, sizeof( reenckey_t ),        "reenckey  :" );
            log_msg_bin( ADEONA_CRYPTO, (uchar*)&cache[0], sizeof( cache_elem_t ), "cache[0]  :" );
            log_msg_bin( ADEONA_CRYPTO, (uchar*)&cache[1], sizeof( cache_elem_t ), "cache[1]  :" );
        }
        
        debug( "=================== end cache update =======================" );
        
        // Move forward cacheseed
		memcpy( cacheseed, keys.nextseed, sizeof(prgkey_t) );
		
        return SUCCESS;
	}
	else {
		error( "invalid input" );
		return FAILURE;
	}

}

/*
 * Note: the cache does not get popped if verify_then_decrypt fails.
 * Not sure I'm clear on whether it should or not.
 *
 */ 
int ctxt_cache_getelem( prgkey_t cacheseed IN, cache_elem_t* cache IN, update_t* update )
{
    if( cacheseed != NULL && cache != NULL && update != NULL ) {
        prgkey_t seed; 
		cache_keys_t keys;
		cache_elem_t elem;
        uchar ptxt[PTXT_SIZE];
		
        memcpy( seed, cacheseed, sizeof( prgkey_t ) );
        
        debug( "================== start cache getelem ======================" );       

        elem = cache[0];    
        // Seek forward to find the correct cacheseed
        int i;
        log_msg_bin( ADEONA_DEBUG, elem.sentinel, SENTINEL_SIZE, "looking for sentinel value" );
        for( i = 0; i < MAX_SENTINEL_TRIES; i++ ) {
            if( prg_generate( seed, 0, sizeof( cache_keys_t ) - 1, (uchar*)&keys ) == FAILURE ) {
			    error( "could not generate keys from cache seed on try %d", i );
			    return FAILURE;
		    }   
            if( memcmp( elem.sentinel, keys.sentinel, SENTINEL_SIZE ) == 0 ) {
                debug( "located correct sentinel, iteration %d", i );       
                break;
            }
		    memcpy( seed, keys.nextseed, sizeof(prgkey_t) );
        }
        if( i == MAX_SENTINEL_TRIES ) {
            error( "couldn't find correct sentinel value" );
            return FAILURE;
        }

        elem = cache[0];    
        
        log_msg_bin( ADEONA_CRYPTO, seed, sizeof( prgkey_t ),                  "cacheseed :" );
        log_msg_bin( ADEONA_CRYPTO, keys.enckey, sizeof( enckey_t ),           "enckey    :" );
        log_msg_bin( ADEONA_CRYPTO, keys.hmackey, sizeof( hmackey_t ),         "hmackey   :" );
        //log_msg_bin( ADEONA_CRYPTO, keys.renckey, sizeof( reenckey_t ),        "reenckey  :" );
        log_msg_bin( ADEONA_CRYPTO, (uchar*)&cache[0], sizeof( cache_elem_t ), "cache[0]  :" );
        
        if( opts.ctxt_queue_size > 1 ) {
            log_msg_bin( ADEONA_CRYPTO, (uchar*)&cache[1], sizeof( cache_elem_t ), "cache[1]  :" );
        }
        
        if( verify_then_decrypt( keys.enckey, 
							     keys.hmackey, 
							     elem.ctxt, CTXT_SIZE, 
							     ptxt, PTXT_SIZE ) == FAILURE ) 
        {
			error( "could not verify and decrypt cache element" );
			return FAILURE;
		}

        // Copy over header, data
        datacpy( &update->skipped, ptxt, sizeof( int ) );
        location_deserialize( ptxt + sizeof(int), LOC_DATA_UPPERBOUND, &update->loc );
        
        // If we have multiple elements, we need to remove re-encryption
		if( opts.ctxt_queue_size > 1 ) {
            
			memmove( &cache[0], &cache[1], CTXT_QUEUE_LEN - sizeof( cache_elem_t ) );
			// dereencrypt (tom's word) old part of cache
			//if( decryption( keys.renckey,
            //                (uchar*)&cache[0],
            //                (uchar*)&cache[0], 
            //                CTXT_QUEUE_LEN - sizeof( cache_elem_t ) ) == FAILURE ) 
            //{
			//	error( "could not dereencrypt ciphertext cache" );
			//	return FAILURE;
			//}
            
            // NOTE: this was causing a segfault for normal bulk retrieves!
            log_msg_bin( ADEONA_CRYPTO, (uchar*)&cache[0], sizeof( cache_elem_t ), "cache[1]->cache[0] after dereencryption:" );
		}
        debug( "=================== end cache getelem =======================" );
        
        
        return SUCCESS;
    }
    else {
		error( "invalid input" );
		return FAILURE;
    }
}


