/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <math.h>
#include <unistd.h>

#include "common.h"
#include "sched.h"
#include "state.h"
#include "util.h"
#include "options.h"
#include "log.h"

/*
 * Get a uniform random number on the unit interval [0,1] 
 * using random <seed>.
 *
 */
static float get_uniform_randvar( unsigned int seed )
{
    float U;
    U = (float)seed;
    U /= UINT_MAX;
    return U;
}

/*
 * Get an exponential variate with parameter <lambda> using the
 * inversion method.
 *
 */ 
static float get_exponential_randvar( unsigned int seed, float lambda )
{
    float U = get_uniform_randvar( seed );
    float T = 0.0f - log( U );
    T /= lambda;
    return T;
}

static int get_delta( state_t* state )
{
    assert( state != NULL && state->sd != NULL );
    
    unsigned int seed;
    datacpy( &seed, state->sd->schedseed, sizeof( unsigned int ) );
    float delta = get_exponential_randvar( seed, opts.sched_lambda );
    
    int idelta = (int)ceilf( delta );
    if( idelta < 1 ) {
        idelta = 1; //sanity check, this should not happen
    }

    return idelta;
}

static int generate_seek_sd( state_t* state )
{
    assert( state != NULL );
    
    if( prg_generate( state->currentseed, STATE_NEXTSEED_LOW, STATE_DSEED_HIGH, (uchar*)state->sd ) == FAILURE ) {
        return FAILURE;
    }
    
    return SUCCESS;
}

/*
 * We call this to generate the rest of the state_data i.e. the parts
 * that didn't need to be generated in order to move the state forward.
 *
 */
static int generate_nonseek_sd( state_t* state )
{
    assert( state != NULL );
    
    if( prg_generate( state->currentseed, STATE_RENC_LOW, STATE_RENC_HIGH, (uchar*)state->sd->renckey ) == FAILURE 
        || prg_generate( state->currentseed, STATE_DHT_LOW, STATE_DHT_HIGH, (uchar*)state->sd->dhtkeys ) == FAILURE ) {
        return FAILURE;
    }
    
    return SUCCESS;
}

/*
 * Note: ONLY generates the parts of the state required to move forward
 * Returns sum of delta skipped if successful and -1 if error
 *
 */
static int seek_state( state_t* state, int nsteps )
{
    assert( state != NULL );    
    
    int i, sumdelta = 0;
    for( i=0; i<nsteps; i++ ) {
        int delta = get_delta( state );
        assert( delta >= 0 );

        state->updatetime = state->updatetime + delta;
        sumdelta += delta;

        memcpy( state->currentseed, state->sd->nextseed, sizeof( prgkey_t ) );
        memcpy( state->cacheseed, state->sd->nextcacheseed, sizeof( prgkey_t ) );
        if( generate_seek_sd( state ) == FAILURE ) {
            error( "failed to generate seed to go forward" );
            return -1;
        }
    }
    
    return sumdelta;
}

int sched_nextstate( state_t* state INOUT )
{
    if( state != NULL && state->sd != NULL ) {

        if( seek_state( state, 1 ) == -1 ) {
            return FAILURE;
        }
        
        return generate_nonseek_sd( state );
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Goes back n states
 *
 * If init is less than n states away from curr, then 
 * curr is rewound to init, and FAILURE is returned
 *
 */ 
int sched_gobackn( state_t* init IN, state_t* curr INOUT, int n IN ) 
{
    if( init != NULL && curr != NULL && n > 0 ) {

        state_t* temp = state_alloc();
        state_t* nback = state_alloc();

        state_copy( nback, init, SC_NOCACHE );
        state_copy( temp, init, SC_NOCACHE );

        int i = 0;
        while( temp->updatetime < curr->updatetime ) {
            seek_state( temp, 1 );
            if( i++ >= n ) {
                seek_state( nback, 1 );
            }
        }
        
        state_copy( curr, nback, SC_NOCACHE );
        state_free( temp );
        state_free( nback );
        
        //generate the rest of the state's data
        if( generate_nonseek_sd( curr ) == FAILURE ) {
            error( "failed to generate state's data" );
            return FAILURE;
        }
        
        if( i < n ) {
            error( "could not go back n" );
            return FAILURE;
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Previous state getter
 *
 */ 
int sched_prevstate( state_t* init IN, state_t* toprev INOUT )
{
    if( init != NULL && toprev != NULL) {
        
        state_t* prev = state_alloc();
        
        state_copy( prev, init, SC_NOCACHE );
        int delta = get_delta( prev );
        
        while( ( prev->updatetime + delta  ) < toprev->updatetime ) {
            seek_state( prev, 1 );
            delta = get_delta( prev );
        }
        
        state_copy( toprev, prev, SC_NOCACHE );
        state_free( prev );
        
        //generate the rest of state data
        if( generate_nonseek_sd( toprev ) == FAILURE ) {
            error( "failed to generate state's data" );
            return FAILURE;
        }
        
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

int sched_catchup( state_t* state INOUT, int* skipped OUT )
{
    if( state != NULL && skipped != NULL ) {
        int i = 0, sumdelta = 0;
        
        time_t currtime = time( NULL );
        while( currtime > state->updatetime ) {
            int delta = seek_state( state, 1 ); 
            
            sumdelta += delta;
            i += 1;
        }
        
        debug( "skipped %d states for delta sum of %d", i, sumdelta );
        
        *skipped = i;
        return generate_nonseek_sd( state );
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Returns first state found with updatetime greater than 
 * or equal to time
 * 
 */
int sched_seekpasttime( state_t* state INOUT, time_t time IN )
{
    if( state != NULL ) {
        // Seek through states to find the very first state which
        // has updatetime greater than or equal to time
        while( state->updatetime <= time ) {
            
            if( seek_state( state, 1 ) < 0 ) {
                return FAILURE;
            }
        } 

        return generate_nonseek_sd( state );
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Returns last state found with updatetime less than 
 * or equal to time
 */
int sched_seektotime( state_t* state INOUT, time_t time IN )
{
    if( state != NULL && time > 0) {
        state_t* prevstate = state_alloc();
        int flag = 0;
        while( state->updatetime < time ) {
            
            state_copy( prevstate, state, SC_NOCACHE );
            flag = 1;
            if( seek_state( state, 1 ) < 0 ) {
                state_free( prevstate );
                return FAILURE;
            }
        } 

        if( flag ) {
            state_copy( state, prevstate, SC_NOCACHE );
        }

        state_free( prevstate );

        return generate_nonseek_sd( state );
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}


int sched_sleep( state_t* state IN )
{
    if( state != NULL && state->sd != NULL ) {
        time_t currtime = time( NULL );

        unsigned int sleeptime = (unsigned int)( state->updatetime - currtime );

		// Sanity check in case sleeptime is negative
		if( state->updatetime > currtime ) {

            if( sleeptime > opts.max_sleeptime ) {
                
                debug( "sleeping for %u seconds", opts.max_sleeptime );
                fflush( NULL ); //flush buffers before going to sleep
                sync();
                sleep( opts.max_sleeptime );
                return CONTINUE;
            }
            
            debug( "sleeping for %u seconds", sleeptime );
            fflush( NULL ); //flush buffers before going to sleep
            sync();
        
            sleep( sleeptime );
        }
        
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}


