/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */

#include "auth_encrypt.h"
#include "log.h"

/*
 * IMPORTANT: Routines in this file assume one-time keys.
 *
 */ 


/*
 * Perform standard Encrypt-then-MAC mode.
 * ct must be non-null
 * clen contains size of buffer ct points to
 * clen must be enough to contain ciphertext and hmac
 * 
 * <pt> and <ct> can be the same
 *
 */
int encrypt_then_mac( 
        enckey_t ekey IN, 
        hmackey_t hkey IN,
        uchar* pt IN, 
        int plen, 
        uchar* ct OUT, 
        int clen IN )
{
    if( ekey != NULL && hkey != NULL 
        && pt != NULL && plen > 0 
        && ct != NULL  && clen >= plen + HMAC_LEN )
    {
        if( encryption( ekey, pt, ct, plen ) == FAILURE ) {
            error( "could not encrypt" );
            return FAILURE;
        }

        if( hmac_compute( hkey, ct, plen, &ct[plen] ) == FAILURE ) {
            error( "could not hmac" );    
            return FAILURE;
        }

        return SUCCESS;

    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Perform standard Encrypt-then-MAC mode.
 * Caller must deallocate <ct> and it's length is written to <clen>
 *
 * Always allocates!
 *
 * <pt> and <ct> cannot be the same.
 *
 */
int encrypt_then_mac_alloc( 
        enckey_t ekey IN, 
        hmackey_t hkey IN,
        uchar* pt IN, 
        int plen, 
        uchar** ct OUT, 
        int* clen OUT )
{
    if( ekey != NULL && hkey != NULL 
        && pt != NULL && plen > 0 && ct != NULL && clen != NULL && pt != *ct )
    {
        *clen = plen + HMAC_LEN;
        
        *ct = (uchar*)malloc( *clen );
        assert( *ct != NULL );
        
        return encrypt_then_mac( ekey, hkey, pt, plen, *ct, *clen );
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}



/*
 * Verify the MAC assumed to be the last HMAC_LEN bytes of <ct> then
 * perform decryption. 
 * <pt> must be allocated already, its allocated length is input by
 * plen. 
 *
 * <pt> and <ct> can be the same
 *
 */
int verify_then_decrypt(
        enckey_t ekey IN,
        hmackey_t hkey IN,
        uchar* ct IN,
        int clen IN,
        uchar* pt OUT,
        int plen IN )
{
    if( ekey != NULL && hkey != NULL 
        && ct != NULL && clen > 0 
        && pt != NULL && plen >= clen - HMAC_LEN )
    {
        
        if( hmac_verify( hkey, ct, plen, &ct[plen] ) == FAILURE ) {
            warn( "could not verify" );
            return FAILURE;
        }

        if( decryption( ekey, ct, pt, plen ) == FAILURE ) {
            error( "could not decrypt" );
            return FAILURE;
        }

        return SUCCESS;

    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Verify the MAC assumed to be the last HMAC_LEN bytes of <ct> then
 * perform decryption.  Caller must deallocate <pt> and it's length is
 * written to <plen>
 *
 * Always allocates!
 *
 * <pt> and <ct> cannot be the same
 *
 */
int verify_then_decrypt_alloc(
        enckey_t ekey IN,
        hmackey_t hkey IN,
        uchar* ct IN,
        int clen IN,
        uchar** pt OUT,
        int* plen OUT )
{
    if( ekey != NULL && hkey != NULL 
        && ct != NULL && clen > 0 && pt != NULL && plen != NULL && *pt != ct )
    {
        *plen = clen - HMAC_LEN;
        *pt = (uchar*)malloc( *plen );
        assert( *pt != NULL );
        
        if( verify_then_decrypt( ekey, hkey, ct, clen, *pt, *plen ) == FAILURE ) {
            memset( *pt, 0x00, *plen );
            free( *pt );
            *pt = NULL;
            *plen = 0;
            return FAILURE;
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}


