/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include "common.h"
#include "util.h"
#include "prg.h"
#include "hmac.h"
#include "encrypt.h"
#include "auth_encrypt.h"
#include "pbkdf.h"
#include "log.h"

//#include "auth_encrypt.h"

static void test_prg( void )
{
    //test vectors from java implementation
    prgkey_t key1 =    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00";
    prgkey_t test1_1 = "\x66\xE9\x4B\xD4\xEF\x8A\x2C\x3B\x88\x4C\xFA\x59\xCA\x34\x2B\x2E"; 
    prgkey_t test1_2 = "\x47\x71\x18\x16\xE9\x1D\x6F\xF0\x59\xBB\xBF\x2B\xF5\x8E\x0F\xD3";
    
    prgkey_t key2 =    "\x66\xE9\x4B\xD4\xEF\x8A\x2C\x3B\x88\x4C\xFA\x59\xCA\x34\x2B\x2E";
    prgkey_t test2_1 = "\xAB\x4B\x36\xBD\xC8\x1D\x8B\xD2\xBC\xED\x4B\x05\xA7\x08\xB4\x22";
    prgkey_t test2_2 = "\xC2\x16\x7E\x7A\xE6\xF0\xF3\x66\xEC\x26\x45\x07\x75\x78\x3D\xDA";
    
    uchar buffer1_1[PRG_BLOCKSIZE];
    uchar buffer1_2[PRG_BLOCKSIZE];
    uchar buffer1[2*PRG_BLOCKSIZE];
    uchar buffer2_1[PRG_BLOCKSIZE];
    uchar buffer2_2[PRG_BLOCKSIZE];
    uchar buffer2[2*PRG_BLOCKSIZE];
    
    prg_generate( key1, 0, PRG_BLOCKSIZE-1, buffer1_1 );
    prg_generate( key1, PRG_BLOCKSIZE, 2*PRG_BLOCKSIZE-1, buffer1_2 );
    prg_generate( key1, 0, 2*PRG_BLOCKSIZE-1, buffer1 );
    
    prg_generate( key2, 0, PRG_BLOCKSIZE-1, buffer2_1 );
    prg_generate( key2, PRG_BLOCKSIZE, 2*PRG_BLOCKSIZE-1, buffer2_2 );
    prg_generate( key2, 0, 2*PRG_BLOCKSIZE-1, buffer2 );
    
    int i;
    for( i=0; i<PRG_BLOCKSIZE-1; i++ ) {
        assert( test1_1[i] == buffer1_1[i] );
        assert( test1_2[i] == buffer1_2[i] );
        assert( test2_1[i] == buffer2_1[i] );
        assert( test2_2[i] == buffer2_2[i] );
        
        assert( test1_1[i] == buffer1[i] );
        assert( test2_1[i] == buffer2[i] );
    }

    for( i=PRG_BLOCKSIZE; i<2*PRG_BLOCKSIZE-1; i++ ) {
        assert( test1_2[i-PRG_BLOCKSIZE] == buffer1[i] );
        assert( test2_2[i-PRG_BLOCKSIZE] == buffer2[i] );
    }

    printout( "***** %s passed\n", __FUNCTION__ );
}

static void test_hmac( void )
{
    char* msg1 = "I want to hmac this string.";
    char* msg2 = "Another string I want to hmac.";
    int mlen1 = strlen( msg1 ) + 1;
    int mlen2 = strlen( msg2 ) + 1;

    hmackey_t hkey1;
    hmackey_t hkey2;
    get_random_bytes( hkey1, sizeof( hmackey_t ) );
    get_random_bytes( hkey2, sizeof( hmackey_t ) );
    
    uchar hmac1[HMAC_LEN];
    uchar hmac2[HMAC_LEN];
    memset( hmac1, 0x00, HMAC_LEN );
    memset( hmac2, 0x00, HMAC_LEN );
    
    assert( hmac_compute( hkey1, msg1, mlen1, hmac1 ) == SUCCESS );
    assert( hmac_compute( hkey2, msg2, mlen2, hmac2 ) == SUCCESS );

    assert( hmac_verify( hkey1, msg1, mlen1, hmac1 ) == SUCCESS );
    assert( hmac_verify( hkey2, msg2, mlen2, hmac2 ) == SUCCESS );

    assert( hmac_verify( hkey2, msg1, mlen1, hmac1 ) == FAILURE );
    assert( hmac_verify( hkey1, msg2, mlen2, hmac2 ) == FAILURE );
    
    assert( hmac_verify( hkey1, msg2, mlen2, hmac1 ) == FAILURE );
    assert( hmac_verify( hkey2, msg1, mlen1, hmac2 ) == FAILURE );
    
    assert( hmac_verify( hkey1, msg1, mlen1, hmac2 ) == FAILURE );
    assert( hmac_verify( hkey2, msg2, mlen2, hmac1 ) == FAILURE );
    
    printout( "***** %s passed\n", __FUNCTION__ );
}

static void test_encrypt( void )
{
    char* plaintext1 = "I want to encrypt this string";
    char* plaintext2 = "Another string I want to encrypt";
    int plen1 = strlen( plaintext1 ) + 1;
    int plen2 = strlen( plaintext2 ) + 1;

    enckey_t ekey1;
    enckey_t ekey2;
    get_random_bytes( ekey1, sizeof( enckey_t ) );
    get_random_bytes( ekey2, sizeof( enckey_t ) );
    
    uchar ciphertext1[plen1];
    memset( ciphertext1, 0x00, plen1 );
    uchar ciphertext2[plen2];
    memset( ciphertext2, 0x00, plen2 );
    
    uchar plaintxt1[plen1];
    memset( plaintxt1, 0x00, plen1 );
    uchar plaintxt2[plen2];
    memset( plaintxt2, 0x00, plen2 );
    
    assert( encryption( ekey1, plaintext1, ciphertext1, plen1 ) == SUCCESS );
    assert( encryption( ekey2, plaintext2, ciphertext2, plen2 ) == SUCCESS );
    
    assert( decryption( ekey1, ciphertext1, plaintxt1, plen1 ) == SUCCESS );
    assert( decryption( ekey2, ciphertext2, plaintxt2, plen2 ) == SUCCESS );

    assert( memcmp( plaintext1, plaintxt1, plen1 ) == 0 );
    assert( memcmp( plaintext2, plaintxt2, plen2 ) == 0 );
    
    assert( decryption( ekey2, ciphertext1, plaintxt1, plen1 ) == SUCCESS );
    assert( decryption( ekey1, ciphertext2, plaintxt2, plen2 ) == SUCCESS );
    
    assert( memcmp( plaintext1, plaintxt1, plen1 ) != 0 );
    assert( memcmp( plaintext2, plaintxt2, plen2 ) != 0 );

    printout( "***** %s passed\n", __FUNCTION__ );
}

static void test_authencryption( void )
{
    char* plaintext1 = "I want to encrypt this string";
    int plen1 = strlen( plaintext1 ) + 1;

    enckey_t ekey1;
    enckey_t ekey2;
    get_random_bytes( ekey1, sizeof( enckey_t ) );
    get_random_bytes( ekey2, sizeof( enckey_t ) );
    
    hmackey_t hkey1;
    hmackey_t hkey2;
    get_random_bytes( hkey1, sizeof( hmackey_t ) );
    get_random_bytes( hkey2, sizeof( hmackey_t ) );
    
    uchar* ciphertext1 = NULL;
    int clen1 = 0;

    uchar* plaintxt1 = NULL;
    int ptlen1 = 0;

    assert( encrypt_then_mac_alloc( ekey1, hkey1, plaintext1, plen1, &ciphertext1, &clen1 ) == SUCCESS );
    assert( ciphertext1 != NULL && clen1 == ( plen1 + HMAC_LEN ) );

    assert( verify_then_decrypt_alloc( ekey1, hkey1, ciphertext1, clen1, &plaintxt1, &ptlen1 ) == SUCCESS );
    assert( plen1 == ptlen1 );
    assert( memcmp( plaintext1, plaintxt1, plen1 ) == 0 );

    free( plaintxt1 );
    ptlen1 = 0;
    
    assert( verify_then_decrypt_alloc( ekey1, hkey2, ciphertext1, clen1, &plaintxt1, &ptlen1 ) == FAILURE );
    assert( plaintxt1 == NULL && ptlen1 == 0 );
    ptlen1 = 0;

    free( plaintxt1 );
    ptlen1 = 0;
    
    assert( verify_then_decrypt_alloc( ekey2, hkey1, ciphertext1, clen1, &plaintxt1, &ptlen1 ) == SUCCESS );
    assert( plen1 == ptlen1 );
    assert( memcmp( plaintext1, plaintxt1, plen1 ) != 0 ); //verify but not decrypt correctly
    
    free( plaintxt1 );
    ptlen1 = 0;
    printout( "***** %s passed\n", __FUNCTION__ );
}

static inline void do_pbkdf_test( char* testv, int count, int dklen )
{
    char* password = "password";
    char* salt = "ATHENA.MIT.EDUraeburn";
        
    uchar dkey[dklen];
    uchar expected[dklen];
    
    memset( dkey, 0x00, sizeof( dkey ) );
    memset( expected, 0x00, sizeof( expected ) );
    assert( sizeof( dkey ) == sizeof( expected ) );
    assert( memcmp( dkey, expected, sizeof( dkey ) ) == 0 );
    
    int tvlen = strlen( testv );
    assert( 2*dklen == tvlen && 2*sizeof( expected ) == tvlen );
    
    hex2bin( testv, tvlen, expected, sizeof( expected ) );
    pbkdf2_hmacsha1( password, strlen( password ), salt, strlen( salt ), count, dkey, dklen );
    
    assert( sizeof( dkey ) == sizeof( expected ) );
    assert( memcmp( dkey, expected, sizeof( dkey ) ) == 0 );
    
}

/*
 *
 * test vectors from page 9 of RFC 3962
 *
 */
static void test_pbkdf( void )
{
    do_pbkdf_test( "CDEDB5281BB2F801565A1122B2563515", 1, 16 ); 
    do_pbkdf_test( "CDEDB5281BB2F801565A1122B25635150AD1F7A04BB9F3A333ECC0E2E1F70837", 1, 32 ); 

    do_pbkdf_test( "01DBEE7F4A9E243E988B62C73CDA935D", 2, 16 );
    do_pbkdf_test( "01DBEE7F4A9E243E988B62C73CDA935DA05378B93244EC8F48A99E61AD799D86", 2, 32 );

    do_pbkdf_test( "5C08EB61FDF71E4E4EC3CF6BA1F5512B", 1200, 16 );
    do_pbkdf_test( "5C08EB61FDF71E4E4EC3CF6BA1F5512BA7E52DDBC5E5142F708A31E2E62B1E13", 1200, 32 );
    
    printout( "***** %s passed\n", __FUNCTION__ );
}

int main( void )
{
    test_prg();
    test_hmac();
    test_encrypt();
    test_authencryption();
    test_pbkdf();

    exit( EXIT_SUCCESS );
}
