/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */

#include <arpa/inet.h>

#include "gateway_prot.h"
#include "opendht.h"
#include "log.h"

#define OPENDHT_APP_STRING "adeona.cs.washington.edu"
#define OPENDHT_CLIB_STRING "rpcgen"

#define OPENDHT_TTL 604800 // seconds in one week
#define OPENDHT_TIMEOUT 15 // seconds
#define OPENDHT_MAX_NUMRET INT_MAX // max number of values returned on a single get  
#define OPENDHT_MAX_PLACEMARKLEN 100

static CLIENT* _dhtclient = NULL;

static int is_timeout_error( void ) 
{
    assert( _dhtclient != NULL );
    
    struct rpc_err re;
    clnt_geterr( _dhtclient, &re );
    if( re.re_status == RPC_TIMEDOUT ) {
        debug( "dht client rpc timed out" );
        return DHT_TIMEOUT;
    }
    else {
        return DHT_FAILURE;
    }
}

int opendht_connect( struct in_addr serverip IN, int port IN )
{
    if( _dhtclient != NULL ) {
        warn( "client exists, destroying old client" );
        clnt_destroy( _dhtclient );
        _dhtclient = NULL;
    }
    
    struct sockaddr_in sin;
    memset( &sin, 0x00, sizeof( struct sockaddr_in ) );
    sin.sin_family = AF_INET;
    sin.sin_port = htons( port );
    sin.sin_addr = serverip;

    int sockp = RPC_ANYSOCK;
    _dhtclient = clnttcp_create( &sin, BAMBOO_DHT_GATEWAY_PROGRAM,
            BAMBOO_DHT_GATEWAY_VERSION, &sockp, 0, 0 );

    if( _dhtclient == NULL ) {
        return DHT_FAILURE;
    }
    
    //set timeout
    struct timeval tv;
    tv.tv_sec = OPENDHT_TIMEOUT;
    tv.tv_usec = 0;
    if( clnt_control( _dhtclient, CLSET_TIMEOUT, (char*)&tv ) == 0 ) {
        warn( "cannot set timeout value" ); 
    }
    
    return DHT_SUCCESS;
}

int opendht_disconnect( void )
{
    if( _dhtclient == NULL ) {
        warn( "no client" );
    }
    else {
        clnt_destroy( _dhtclient );
        _dhtclient = NULL;
    }
    
    return DHT_SUCCESS;
}

int opendht_isconnected( void )
{
    return ( ( _dhtclient != NULL ) ? TRUE:FALSE );
}

/*
 * Allows inserts up to OPENDHT_MAX_PUTLEN bytes.
 *
 */ 
int opendht_put( dhtkey_t key IN, uchar* value IN, int len IN )
{
    if( _dhtclient == NULL ) {
        error( "no client" );
        return DHT_FAILURE;
    }
    
    if( len > OPENDHT_MAX_PUTLEN ) {
        error( "cannot insert more than %d bytes", OPENDHT_MAX_PUTLEN );
        return DHT_FAILURE;
    }

    if( key != NULL && value != NULL && len > 0 ) {
        bamboo_put_args bpa;
        bamboo_stat* bs;

        memset( &bpa, 0x00, sizeof( bamboo_put_args ) );
        memcpy( bpa.key, key, sizeof( dhtkey_t ) );
        bpa.value.bamboo_value_val = (char*)value;
        bpa.value.bamboo_value_len = len;
        bpa.application = OPENDHT_APP_STRING;
        bpa.client_library = OPENDHT_CLIB_STRING;
        bpa.ttl_sec = OPENDHT_TTL;
        
        bs = bamboo_dht_proc_put_2( &bpa, _dhtclient );
        if( bs == NULL ) {
            return is_timeout_error();
        }
        
        switch( *bs ) {
            case BAMBOO_AGAIN : /* fall through */
                debug( "opendht put return code BAMBOO_AGAIN" );
                return is_timeout_error();
            case BAMBOO_CAP : 
                debug( "opendht put return code BAMBOO_CAP" );
                return is_timeout_error();
            case BAMBOO_OK :
                return DHT_SUCCESS;
            default :
                warn( "unknown put request return code" );
                return is_timeout_error();
        }
    }
    else {
        error( "invalid input" );
        return DHT_FAILURE;
    }
}


/*
 * Attempts to retrieve all values under <key>.
 * Caller must deallocate <result>.
 * 
 * IMPORTANT:
 * Please set <*result> = NULL and <*rlen> = 0 explicitly for an
 * initial call. Otherwise, the routine might use a stale placemark
 * value.
 * 
 * static buffers are used for the data placemark so we don't have to
 * pass pointers around.
 * 
 */ 
int opendht_get( dhtkey_t key IN, uchar** result OUT, unsigned int* rlen INOUT )
{
    if( _dhtclient == NULL ) {
        error( "no client" );
        return DHT_FAILURE;
    }   

    if( key != NULL && result != NULL && rlen != NULL ) {
        static char _placemark[OPENDHT_MAX_PLACEMARKLEN];
        static uint _plen = 0;
        
        // i think simply _plen = 0 would work but i'll just do this
        if( *result == NULL && *rlen  == 0 ) {
            memset( _placemark, 0x00, OPENDHT_MAX_PLACEMARKLEN );
            _plen = 0;
        }

        bamboo_get_args bga;
        bamboo_get_res* bgr;
        bamboo_value* bv;

        memset( &bga, 0x00, sizeof( bamboo_get_args ) );
        memcpy( bga.key, key, sizeof( dhtkey_t ) ); 
        
        bga.placemark.bamboo_placemark_val = _placemark;
        bga.placemark.bamboo_placemark_len = _plen;
        bga.application = OPENDHT_APP_STRING;
        bga.client_library = OPENDHT_CLIB_STRING;
        bga.maxvals = OPENDHT_MAX_NUMRET;
        
        //if the loop fails while there are still values under the
        //<key>, the placemark would have been saved and a subsequent
        //call would just append to <result> 
        do {
            bgr = bamboo_dht_proc_get_2( &bga, _dhtclient );
            if( bgr == NULL ) {
                error( "no get result" );
                return is_timeout_error();   
            }
            
            if( bgr->values.values_len <= 0 ) {
                debug( "successful dht rpc call but no value under key, result len == %d", bgr->values.values_len );
            }

            int i;
            for( i=0; i<bgr->values.values_len; i++ ) {
                bv = &bgr->values.values_val[i];
                *result = (uchar*)realloc( *result, *rlen + bv->bamboo_value_len );
                assert( *result != NULL );
                memcpy( *result + *rlen , bv->bamboo_value_val, bv->bamboo_value_len );
                *rlen += bv->bamboo_value_len;
            }
            
            bga.placemark.bamboo_placemark_len = _plen = bgr->placemark.bamboo_placemark_len;
            memcpy( _placemark, bgr->placemark.bamboo_placemark_val, _plen );

        } while( bgr->placemark.bamboo_placemark_val != NULL );
        
        return DHT_SUCCESS;
    }
    else {
        error( "invalid input" );
        return DHT_FAILURE;
    }
}

